#include "stdafx.h"





namespace ThinAudioGateway {





// #############################################################################################
//
// TagDriver class implementation
// =============================================================================================
// 
// #############################################################################################





//
// Constructor
// =============================================================================================
// 

TagDriver::TagDriver (GUID const & xClassId)

: ClassId (xClassId)

#if _Debug

  , Dbg (&::Dbg, 6, L"Drv")

#endif // _Debug

{

  *InterfacePath = L'\0';

  Debug (6, L"Constructed, class ID %s", GuidText (ClassId));

  Check ();

}





//
// Destructor
// =============================================================================================
// 

TagDriver::~TagDriver () {

  Debug (6, L"Destructing");

  Check ();

  Close ();

}





//
// Find KS filter interface
// =============================================================================================
//

HRESULT TagDriver::FindInterface () {

  Check ();

  HDEVINFO DevList = 0;

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    Debug (7, L"Querying interfaces in class %s", GuidText (ClassId));

    DevList = SetupDiGetClassDevs (
      &ClassId, nullptr, 0,
      DIGCF_DEVICEINTERFACE | DIGCF_PROFILE | DIGCF_PRESENT
    );

    if (DevList == INVALID_HANDLE_VALUE) {

      Res = HRESULT_FROM_SETUPAPI (GetLastError ());

      Debug (9, L"Cannot get class %s interfaces (%s)", GuidText (ClassId), ErrorText (Res));

      break;

    }

    SP_DEVICE_INTERFACE_DATA IfData;

    ZeroMemory (&IfData, sizeof (IfData));

    IfData.cbSize = sizeof (IfData);

    if (!SetupDiEnumDeviceInterfaces (DevList, nullptr, &ClassId, 0, &IfData)) {

      Res = HRESULT_FROM_SETUPAPI (GetLastError ());

      if (Res == HRESULT_FROM_WIN32 (ERROR_NO_MORE_ITEMS)) {

        Res = HRESULT_FROM_WIN32 (ERROR_NOT_FOUND);

        Debug (9, L"No interfaces found");

      } else {

        Debug (9, L"Cannot get first interface info (%s)", ErrorText (Res));

      }

      break;

    }

    if (!(IfData.Flags & SPINT_ACTIVE)) {

      Debug (6, L"Interface is disabled");

      break;

    }

    struct IfDetail : public SP_DEVICE_INTERFACE_DETAIL_DATA {

      wchar_t VarPart [MAX_PATH];
      
    } IfDetData;

    ZeroMemory (&IfDetData, sizeof (IfDetData));
    IfDetData.cbSize = sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);

    DWORD ReqSize;

    SP_DEVINFO_DATA DevInfo;

    ZeroMemory (&DevInfo, sizeof (DevInfo));
    DevInfo.cbSize = sizeof (DevInfo);

    if (
      !SetupDiGetDeviceInterfaceDetail (
        DevList,
        &IfData,
        &IfDetData,
        sizeof (IfDetData),
        &ReqSize,
        &DevInfo
      )
    ) {

      Res = HRESULT_FROM_SETUPAPI (GetLastError ());

      Debug (9, L"Cannot get interface details (%s)", ErrorText (Res));

      break;

    }

    LocSize const IfPathLen = static_cast <LocSize> (wcslen (IfDetData.DevicePath));

    if (IfPathLen >= NumArrayElems (InterfacePath)) {

      Debug (9, L"Interface path too long (%u)", IfPathLen);

      Res = HRESULT_FROM_WIN32 (ERROR_BUFFER_OVERFLOW);

      break;

    }

    wcscpy (InterfacePath, IfDetData.DevicePath);

    Res = S_OK;

  } OnlyOnce;

  if (DevList != 0) {

    SetupDiDestroyDeviceInfoList (DevList);
    DevList = 0;

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





//
// Create driver object
// =============================================================================================
// 

HRESULT _stdcall TagDriver::Create (TagDriver * & Obj, GUID const & xClassId) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  TagDriver * Drv = nullptr;

  Obj = nullptr;

  do {

#if _Debug

    ::Dbg.Printf (7, L"Creating driver object");

#endif _Debug

    Drv = new TagDriver (xClassId);

    if (Drv == nullptr) {

#if _Debug

      ::Dbg.Printf (9, L"No memory for driver object");

#endif _Debug

      Res = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);
      break;

    }

#if _Debug

    ::Dbg.Printf (8, L"Created driver object %p", Drv);

#endif _Debug

    Obj = Drv;

    Obj->Check ();

    Res = S_OK;

  } OnlyOnce;

  if (FAILED (Res)) {

    if (Drv != nullptr) {

      delete Drv;
      Drv = nullptr;

    }

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Delete driver object
// =============================================================================================
// 

void TagDriver::Delete () {

  Debug (8, L"Deleting driver object %p", this);

  delete this;

}





//
// Open the driver
// =============================================================================================
// 

HRESULT TagDriver::Open () {

  Check ();

  Assert (!IsOpen ());

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    Res = FindInterface ();

    if (FAILED (Res)) {

      break;

    }

    Assert (IsOpen ());

    Res = S_OK;

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





//
// Close the driver
// =============================================================================================
// 

void TagDriver::Close () {

  Check ();

  *InterfacePath = L'\0';

  Check ();

}





HRESULT TagDriver::CreatePipe (TagPipe * & Pipe, bool const Capture) {

  Check ();

  Assert (IsOpen ());

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  TagPipe * P = nullptr;

  do {

    P = new TagPipe (this, Capture);

    if (P == nullptr) {

      Debug (9, L"No memory for %s pipe", P->GetType ());

      Res = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);

      break;

    }

    Res = P->Open (InterfacePath);

    if (FAILED (Res)) {

      break;

    }

    Pipe = P;

    Assert (SUCCEEDED (Res));

  } OnlyOnce;

  if (FAILED (Res)) {

    if (P != nullptr) {

      delete P;
      P = nullptr;

    }

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





} // namespace
