#include "stdafx.h"





namespace ThinAudioGateway {





// #############################################################################################
//
// Constants
// =============================================================================================
// 
// #############################################################################################





static PCTSTR const KsStateNames [] = {L"Stop", L"Acquire", L"Pause", L"Run" };





// #############################################################################################
//
// TagHost class implementation
// =============================================================================================
// 
// #############################################################################################





//
// Constructor
// =============================================================================================
// 

TagHost::TagHost (TagDriver & Drv, bool const Capt, CallbackInterface & CB)

: Driver (Drv), Capture (Capt), Callback (CB),

  Event (0),
  Pipe (nullptr),
  Started (false),
  EventLoopActive (false),
  ExitRequest (false)

#if _Debug

  , Dbg (&::Dbg, 6, L"Host %c", IsCapt ()? L'c' : L'r')

#endif // _Debug

{

}





//
// Destructor
// =============================================================================================
//

TagHost::~TagHost () {

  Stop ();

}





//
// Get common data address
// =============================================================================================
// 

CommonDataDesc const & TagHost::GetCommonData () const {

  Assert (Pipe != nullptr);

  return Pipe->GetCommonData ();

}





//
// Create object instance
// =============================================================================================
// 

HRESULT _stdcall TagHost::Create (TagHost * & Obj, TagDriver & Drv, bool const Capture, CallbackInterface & CB) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  TagHost * Host = nullptr;

  Obj = nullptr;

  do {

#if _Debug

    ::Dbg.Printf (7, L"Creating host object");

#endif _Debug

    Host = new TagHost (Drv, Capture, CB);

    if (Host == nullptr) {

#if _Debug

      ::Dbg.Printf (9, L"No memory for host object");

#endif _Debug

      Res = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);
      break;

    }

#if _Debug

    ::Dbg.Printf (8, L"Created host object %p", Host);

#endif _Debug

    Obj = Host;

    Obj->Check ();

    Res = S_OK;

  } OnlyOnce;

  if (FAILED (Res)) {

    if (Host != nullptr) {

      delete Host;
      Host = nullptr;

    }

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Delete object instance
// =============================================================================================
// 

void TagHost::Delete () {

  Debug (8, L"Deleting host object %p", this);

  delete this;

}





//
// Start streaming
// =============================================================================================
//

HRESULT TagHost::Start (WaveFormat const & Format, UINT const MinBufMs) {

  Assert (Event == 0 && Pipe == nullptr && !Started && !EventLoopActive && !ExitRequest);

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    Res = Driver.CreatePipe (Pipe, IsCapt ());

    if (FAILED (Res)) {

      Debug (9, L"Cannot obtain pipe (%s)", ErrorText (Res));

      break;

    }

    Event = CreateEvent (nullptr, false, false, nullptr);

    if (Event == 0) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      Debug (9, L"Cannot create notification event (%s)", ErrorText (Res));

      break;

    }

    Res = Pipe->SetNotificationEvent (Event);

    if (FAILED (Res)) {

      Debug (9, L"Cannot set notification event (%s)", ErrorText (Res));

      break;

    }

    Res = Pipe->SetFormat (Format.GetSR (), Format.GetBPS (), Format.GetNC (), Format.GetBPC (), Format.GetCC ());

    if (FAILED (Res)) {

      Debug (
        9,
        L"Cannot set format to %u/%u(%u)/%u(%x) (%s)",
        Format.GetSR (), Format.GetBPS (), Format.GetNC (), Format.GetBPC (), Format.GetCC (), ErrorText (Res)
      );

      break;

    }

    Fmt = Format;

    Debug (
      7,
      L"Format set to %u/%u(%u)/%u(%x)",
      Fmt.GetSR (), Fmt.GetBPS (), Fmt.GetNC (), Fmt.GetBPC (), Fmt.GetCC ()
    );

    if (MinBufMs != 0) {

      UINT const Frames = Fmt.GetSR () * MinBufMs / 1000;

      Assert (Frames != 0);

      // If the driver increases buffer size above requested by the Windows Audio Graph
      // (AudioDG), the Graph usually fails its client's request.

      Res = Pipe->SetMinimalBufferSize (Frames);

      if (FAILED (Res)) {

        Debug (9, L"Cannot set minimal buffer size to %u frames (%s)", Frames, ErrorText (Res));

        break;

      }

    }

    Started = true;

    Debug (7, L"Started");

    Res = S_OK;

  } OnlyOnce;

  if (FAILED (Res)) {

    Stop ();

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Stop streaming
// =============================================================================================
//

HRESULT TagHost::Stop () {

  if (EventLoopActive) {

    ExitRequest = true;

    while (EventLoopActive) {

      Sleep (15);

    }

    Assert (!ExitRequest);

  }

  Started = false;

  if (Pipe != nullptr) {

    Pipe->Delete ();
    Pipe = nullptr;

  }

  if (Event != 0) {

    CloseHandle (Event);
    Event = 0;

  }

  return S_OK;

}





//
// Event loop
// =============================================================================================
//

HRESULT TagHost::EventLoop (UINT const PollPeriodMs) {

  Assert (PollPeriodMs != 0);

  Assert (Pipe != nullptr && Pipe->IsOpen () && !EventLoopActive);

  EventLoopActive = true;

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    CommonDataDesc & CD = Pipe->GetCommonData ();
    StreamInfo volatile & SI = CD.Drv.SI;

    PBYTE & Buffer = (PBYTE &) (CD.Host.Buffer);

    UINT BufferFrames = 0;
    UINT FramesInPacket = 0;

    bool HaveStream = false;

    KSSTATE State = KSSTATE_STOP;

    UINT volatile & DevicePos = CD.Host.DevicePos;

    timeBeginPeriod (1);

    LARGE_INTEGER QpcFreq;

    Verify (QueryPerformanceFrequency (&QpcFreq));

    Debug (7, L"QPC frequency: %u", QpcFreq.QuadPart);

    Assert (QpcFreq.QuadPart != 0);

    LARGE_INTEGER LastTimeQpc;

    LastTimeQpc.QuadPart = 0;

    Debug (7, L"Starting event loop");



    for (;;) {

      if (ExitRequest) {

        Debug (6, L"Exit request detected");

        ExitRequest = false;

        Res = S_OK;
        break;

      }

      // Wait for a driver notification or for a timeout.

      DWORD const WaitRes = WaitForSingleObject (Event, PollPeriodMs);

      //Debug (5, L"WR=%u, TRE=%u", WaitRes, ThreadExitRequest);

      if (WaitRes == WAIT_OBJECT_0) {

        //Debug (
        //  7,
        //  L"FI.Flags: %#x, FI.HasStream: %u, SI.State: %u, SI.BufferFrames: %u",
        //  CD.Drv.FI.Flags, CD.Drv.FI.HasStream, SI.State, SI.BufferFrames
        //);

        if (TestBits (CD.Drv.FI.Flags, FilterFlags::WaitingForHost)) {

          // Detect stream creation/termination

          if (!HaveStream && CD.Drv.FI.HasStream) {

            Debug (7, L"Stream created");

            HaveStream = true;
            Assert (State == KSSTATE_STOP);

          }

          if (HaveStream && !CD.Drv.FI.HasStream) {

            Debug (7, L"Stream terminated");

            State = KSSTATE_STOP;
            HaveStream = false;

            Assert (DevicePos == 0);

          }
          


          // Detect buffer creation/deletion

          if (BufferFrames != SI.BufferFrames) {
          
            if (BufferFrames == 0 && SI.BufferFrames != 0) {
            
              Debug (7, L"Buffer acquired (%u frames)", SI.BufferFrames);

            } else if (BufferFrames != 0 && SI.BufferFrames == 0) {
            
              Debug (7, L"Buffer released");

              FramesInPacket = 0;

            } else {

              Debug (6, L"Buffer size changed (%u -> %u frames)", BufferFrames, SI.BufferFrames);

            }

            BufferFrames = SI.BufferFrames;

            Assert (BufferFrames % 2 == 0);

            FramesInPacket = BufferFrames / 2;

          }



          // Detect stream state changement

          if (State != SI.State) {

            Assert (SI.State <= NumArrayElems (KsStateNames));

            if (abs (int (State - SI.State)) > 1) {

              Debug (9, L"Lost some state transitions");

            }

            Debug (6, L"State transition %s -> %s", KsStateNames [State], KsStateNames [SI.State]);

            Debug (5, L"DevPos: %u", DevicePos);

            if (SI.State == KSSTATE_STOP) {

              Assert (State == KSSTATE_ACQUIRE);

              Debug (7, L"Stream stopped");

            } else if (SI.State == KSSTATE_RUN) {

              Assert (State == KSSTATE_PAUSE);

              Debug (7, L"Stream started");

              Verify (QueryPerformanceCounter (&LastTimeQpc));
              Assert (LastTimeQpc.QuadPart != 0);

            } else if (State == KSSTATE_RUN && SI.State == KSSTATE_PAUSE) {

              Debug (7, L"Stream paused");

            }

            State = SI.State;

          }

          // Confirm the driver request

          Res = Pipe->Notify (STATUS_SUCCESS);

          if (FAILED (Res)) {

            break;

          }

        } // Filter/stream state analyzing

        continue; // Nothing to do more after driver's notification

      } // Successful wait actions



      if (/*!Pipe->IsRunningState () ||*/ State != KSSTATE_RUN) {

        continue; // Wait for a next event

      }



      Assert (LastTimeQpc.QuadPart != 0);

      LARGE_INTEGER CurrentTimeQpc;

      Verify (QueryPerformanceCounter (&CurrentTimeQpc));

      Assert (CurrentTimeQpc.QuadPart != 0);

      UINT64 const ElapsedTime = (
        static_cast <UINT64> (CurrentTimeQpc.QuadPart) - static_cast <UINT64> (LastTimeQpc.QuadPart)
      );

      LastTimeQpc.QuadPart = CurrentTimeQpc.QuadPart;

      UINT Frames = static_cast <UINT> ((ElapsedTime * Fmt.GetSR () + QpcFreq.QuadPart / 2) / QpcFreq.QuadPart);

      Debug (5, L"Period: %u fms", Frames);

      Frames = min (Frames, FramesInPacket);

      // Copy data by portions

      while (Frames != 0) {

        Assert (DevicePos < BufferFrames);

        UINT const Portion = min (Frames, BufferFrames - DevicePos);

        Assert (Portion != 0);

        LocSize const Bytes = Fmt.FramesToBytes (Portion);

        LocSize const Actual = Callback.CopyData (Buffer + (DevicePos * Fmt.GetFrameSize ()), Bytes);

        Assert (Actual % Fmt.GetBPC () == 0);

        if (IsCapt ()) {

          Debug (5, L"Read %u fms at %u", Actual, DevicePos);

        } else {

          Debug (5, L"Written %u fms at %u", Actual, DevicePos);

        }

        Assert (DevicePos + Portion <= BufferFrames);

        Pipe->AdvanceDevicePosition (Fmt.BytesToFrames (Actual));

        Assert (Portion <= Frames);

        Frames -= Portion;

      } // Portion read/write loop



    } // Event loop


    Debug (7, L"Event loop ended");

    timeEndPeriod (1);

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  EventLoopActive = false;

  return Res;

}





} // namespace
