#include "stdafx.h"





namespace ThinAudioGateway {





// #############################################################################################
//
// Constants
// =============================================================================================
// 
// #############################################################################################





// #############################################################################################
//
// Variables
// =============================================================================================
// 
// #############################################################################################





// #############################################################################################
//
// TagPipe class implementation
// =============================================================================================
// 
// #############################################################################################





//
// Constructor
// =============================================================================================
// 

TagPipe::TagPipe (TagDriver const * xDrv, bool const C)

: Drv (xDrv), Capture (C)

#if _Debug

  , Dbg (&::Dbg, 6, L"Pipe %c", IsCapt ()? L'c' : L'r')

#endif

{

  Connection = 0;
  FrameSize = 0;
  CommonData = nullptr;

  Debug (6, L"Constructed");

  Check ();

}





//
// Destructor
// =============================================================================================
// 

TagPipe::~TagPipe () {

  Debug (6, L"Destructing");

  Check ();

  Close ();

}





//
// Delete pipe object
// =============================================================================================
// 

void TagPipe::Delete () {

  Check ();

  Debug (8, L"Deleting pipe object %p", this);

  delete this;

}





//
// Open (connect to the driver)
// =============================================================================================
// 

HRESULT TagPipe::Open (PCTSTR const DriverInterface) {

  Assert (DriverInterface != nullptr && *DriverInterface != L'\0');

  Check ();

  Assert (!IsOpen ());

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    wchar_t PipeInterface [MAX_PATH];

    _snwprintf (
      PipeInterface, NumArrayElems (PipeInterface),
      L"%s\\%s\\%s",
      DriverInterface, GuidText (Drv->GetClassId ()),
      IsCapt ()? CaptureConnectionSfx : RenderConnectionSfx
    );

    Debug (7, L"Opening pipe interface %s", PipeInterface);

    Connection = CreateFile (
      PipeInterface, GENERIC_READ | GENERIC_WRITE, 0, nullptr, OPEN_EXISTING, 0, 0
    );

    if (Connection == INVALID_HANDLE_VALUE) {

      Connection = 0;

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      if (Res == HRESULT_FROM_WIN32 (ERROR_FILE_NOT_FOUND)) {

        Debug (9, L"Pipe interface %s not found", PipeInterface);

      } else {

        Debug (9, L"Cannot open pipe interface %s (%s)", PipeInterface, ErrorText (Res));

      }

      break;

    }

    Debug (7, L"Open");

    Assert (CommonData == nullptr);

    GetCommonDataAddressReq Req;

    Res = DriverRequest (
      IoCtl_Local_GetCommonDataAddress, Req, sizeof (Req), &CommonData, sizeof (CommonData)
    );

    if (FAILED (Res)) {

      Debug (9, L"Cannot get common data address (%s)", ErrorText (Res));

      break;

    }

    Debug (7, L"Common data: %Ix:%u", CommonData, sizeof (*CommonData));

    if (IsBadReadPtr (CommonData, sizeof (*CommonData))) {

      Res = HRESULT_FROM_WIN32 (ERROR_ACCESS_DENIED);

      Debug (9, L"Common data block (%Ix:%u) is not readable", CommonData, sizeof (*CommonData));

      break;

    }

    Debug (
      7,
      L"Audio buffer: %Ix:%u",
      CommonData->Host.Buffer, CommonData->Drv.FI.ReservedBufferSize
    );

    if (IsBadReadPtr (CommonData->Host.Buffer, CommonData->Drv.FI.ReservedBufferSize)) {

      Res = HRESULT_FROM_WIN32 (ERROR_ACCESS_DENIED);

      Debug (
        9,
        L"Audio buffer (%Ix:%u) is not readable",
        CommonData->Host.Buffer, CommonData->Drv.FI.ReservedBufferSize
      );

      break;

    }

    Res = S_OK;

  } OnlyOnce;

  if (FAILED (Res)) {

    Close ();

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





//
// Close (disconnect from the driver)
// =============================================================================================
// 

void TagPipe::Close () {

  Check ();

  if (IsOpen ()) {

    CommonData = nullptr;

    CloseHandle (Connection);
    Connection = 0;

    Debug (7, L"Closed");

  }

  Check ();

}





//
// Issue a driver request
// =============================================================================================
// 

HRESULT TagPipe::DriverRequest (
  UINT const Func,
  DriverReq & Req, LocSize const ReqSize,
  void * const Data, LocSize const DataSize,
  LocSize * const RetSize
) const {

  // Bypass checking if not fully initialized

  if (Func != IoCtl_Local_GetCommonDataAddress) {

    Check ();

  }

  Assert (IsOpen ());

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    Req.ApiVer = ApiVersion;

    DWORD BytesRet;

    bool const Success = !!DeviceIoControl (
      Connection, Func,
      &Req, UINT (ReqSize), Data, UINT (DataSize),
      &BytesRet, nullptr
    );

    if (RetSize != nullptr) {

      *RetSize = BytesRet;

    }

    if (!Success) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      Debug (6, L"*** Driver request (%#x) failed (%s)", Func, ErrorText (Res));

      break;

    }

    Res = S_OK;

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Set minimal buffer size for a client
// =============================================================================================
// 

HRESULT TagPipe::SetMinimalBufferSize (UINT const Frames) {

  Assert (Frames != 0);

  Check ();

  Assert (IsOpen ());

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    SetMinimalBufferSizeReq Req;

    Req.Frames = Frames;

    Res = DriverRequest (IoCtl_Local_SetMinimalBufferSize, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot set buffer size (%s)", ErrorText (Res));

      break;

    }

    Debug (7, L"Minimal buffer size set to %u frames", Req.Frames);

    Res = S_OK;

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





//
// Set audio format available for driver's subdevice clients
// =============================================================================================
// 

HRESULT TagPipe::SetFormat (UINT const SR, UINT const BPS, UINT const NC, UINT const BPC, UINT const CC) {

  AssertInRange (SR, SamplingRateMin, SamplingRateMax);
  AssertInRange (BPS, BitsPerSampleMin, BitsPerSampleMax);
  AssertInRange (NC, NumChannelsMin, NumChannelsMax);
  AssertInRange (BPC, BytesPerContainerMin, BytesPerContainerMax);
  Assert (CC == UINT (-1) || !TestBits (CC, SPEAKER_RESERVED));

  Check ();

  Assert (IsOpen ());

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    WaveFormat NewFmt (SR, BPS, NC, BPC, CC);

    SetDataFormatReq Req;

    Req.DataFormat = NewFmt.GetFormat ();

    Assert (Req.DataFormat.SamplingRate == SR && Req.DataFormat.BitsPerSample == BPS);
    Assert (Req.DataFormat.NumChannels == NC && Req.DataFormat.BytesPerContainer == BPC);

    Debug (6, L"Setting filter format to %s", WfxText (NewFmt));

    Res = DriverRequest (IoCtl_Local_SetDataFormat, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot set filter data format (%s)", ErrorText (Res));

      break;

    }

    Fmt = NewFmt;

    Debug (7, L"Filter format set to %s", WfxText (Fmt));

    Res = S_OK;

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





//
// Set a notification event
// =============================================================================================
// 

HRESULT TagPipe::SetNotificationEvent (HANDLE const Event) {

  Assert (Event != 0);

  Check ();

  Assert (IsOpen ());

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    SetNotificationEventReq Req;

    Req.Event = Event;

    Res = DriverRequest (IoCtl_Local_SetNotificationEvent, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot set notification event (%s)", ErrorText (Res));

      break;

    }

    Debug (7, L"Notification event %#x passed to the filter", Event);

    Assert (Res == S_OK);

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





//
// Notify driver about host's state change
// =============================================================================================
//

HRESULT TagPipe::Notify (NTSTATUS const Code) {

  Check ();

  Assert (IsOpen ());

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    Debug (5, L"Notifying driver (%#x)", Code);

    DriverNotificationReq Req;

    Req.Code = Code;

    Res = DriverRequest (IoCtl_Local_Notify, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot notify driver with code %#x (%s)", ErrorText (Res));

      break;

    }

    Assert (Res == S_OK);

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





//
// Reset host/device audio position (read/write buffer offset)
// =============================================================================================
//
// On entering KSSTATE_STOP, position is reset by the filter.
//

HRESULT TagPipe::ResetDevicePosition () {

  Check ();

  Assert (IsOpen ());

  CommonData->Host.DevicePos = 0;
  CommonData->Host.DevicePosBytes = 0;

  Debug (6, L"Device position reset to 0");

  Check ();

  return S_OK;

}





//
// Advance host/device audio position (read/write buffer offset)
// =============================================================================================
// 

HRESULT TagPipe::AdvanceDevicePosition (UINT const Frames) {

  Assert (Frames != 0);

  Check ();

  Assert (IsOpen ());

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    UINT volatile & BufferFrames = CommonData->Drv.SI.BufferFrames;
    UINT volatile & DevPos = CommonData->Host.DevicePos;

    Assert (BufferFrames != 0);

#if _Debug

    UINT const PrevPos = DevPos;

#endif

    // Update main position in frames

    DevPos = (DevPos + Frames) % BufferFrames; // Must be written once

    // Update auxiliary position in bytes (used as a position register for the client)

    CommonData->Host.DevicePosBytes = Fmt.FramesToBytes (DevPos); // Must be written once

    Debug (5, L"Adv dev pos by %u fms (%u->%u)", Frames, PrevPos, DevPos);

    Res = Notify (StatusCodes::CommonDataUpdated);

    if (FAILED (Res)) {

      break;

    }

    Assert (Res == S_OK);

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





} // namespace
