/*

API definitions

*/





// Set 4-byte alignment to have the same alignment in both 32- and 64-bit code

__pragma (pack (push, 4))





namespace ThinAudioGateway {





// #############################################################################################
//
// Macros
// =============================================================================================
//
// #############################################################################################





//
// API version
// =============================================================================================
//

static UINT const ApiVersion = 3;





#define RenderConnectionSfx L"Render"
#define CaptureConnectionSfx L"Capture"



// Global driver parameter registry value names

#define MinSamplingRateValName L"Minimum sampling rate"
#define MaxSamplingRateValName L"Maximum sampling rate"

#define MinBitsPerSampleValName L"Minimum bits per sample"
#define MaxBitsPerSampleValName L"Maximum bits per sample"

#define MinNumChannelsValName L"Minimum number of channels"
#define MaxNumChannelsValName L"Maximum number of channels"





// #############################################################################################
//
// Constants
// =============================================================================================
//
// #############################################################################################




enum IoCtlFuncs : UINT {

  IOCF_DriverFuncStart = 0x800,

  IOCF_SetDriverDataRange,
  IOCF_ResetDriverCounters,

  IOCF_FilterFuncStart = 0x900,

  IOCF_GetCommonDataAddress,
  IOCF_SetDataFormat,
  IOCF_SetMinimalBufferSize,
  IOCF_SetNotificationEvent,
  IOCF_Notify,
  //IOCF_ResetCounters,
  IOCF_SetVolumeControlParams,

};





#define DefineIoctlCode(Name, Access) \
  static UINT const IoCtl_Local_##Name = CTL_CODE ( \
  FILE_DEVICE_UNKNOWN, \
  static_cast <UINT> (IOCF_##Name), \
  METHOD_BUFFERED, \
  Access \
);



DefineIoctlCode (SetDriverDataRange, FILE_WRITE_DATA)
DefineIoctlCode (ResetDriverCounters, FILE_WRITE_DATA)

DefineIoctlCode (GetCommonDataAddress, FILE_READ_DATA)
DefineIoctlCode (Notify, FILE_WRITE_DATA)
DefineIoctlCode (SetDataFormat, FILE_WRITE_DATA)
DefineIoctlCode (SetMinimalBufferSize, FILE_WRITE_DATA)
DefineIoctlCode (SetNotificationEvent, FILE_WRITE_DATA)
//DefineIoctlCode (ResetCounters, FILE_WRITE_DATA)
DefineIoctlCode (SetVolumeControlParams, FILE_WRITE_DATA)





#undef DefineLocalIoctlCode





enum HostNotifications : UINT {

  HN_None,

  HN_StreamCreation,
  HN_StreamDeletion,
  HN_StreamStateChange,
  HN_VolumeChange,

};





// #############################################################################################
//
// Classes
// =============================================================================================
//
// #############################################################################################





//
// Data range descriptor
// =============================================================================================
//

struct DataRangeDesc {

  UINT MinSamplingRate;         // Minimal sampling rate
  UINT MaxSamplingRate;         // Maximal sampling rate
  UINT MinBitsPerSample;        // Minimal bits per sample
  UINT MaxBitsPerSample;        // Maximal bits per sample
  UINT MinNumChannels;          // Minimal number of channels
  UINT MaxNumChannels;          // Maximal number of channels
  UINT MinBytesPerContainer;    // Minimal bytes per container
  UINT MaxBytesPerContainer;    // Maximal bytes per container

};





//
// Data format descriptor
// =============================================================================================
//

struct DataFormatDesc {

  UINT SamplingRate;         // Sampling rate
  UINT BitsPerSample;        // Bits per sample
  UINT NumChannels;          // Number of channels
  UINT BytesPerContainer;    // Bytes per container
  UINT ChannelConfig;        // Channel configuration

  bool operator == (DataFormatDesc const & Other) const { return (memcmp (this, &Other, sizeof (*this)) == 0); }
  bool operator != (DataFormatDesc const & Other) const { return !operator == (Other); }

};





//
// Volume control data descriptor
// =============================================================================================
//

struct VolumeControlParamsDesc {

  bool Enabled;                 // Volume control enabled
  int MinDb;                    // Min level (dB)
  int MaxDb;                    // Max level (dB)
  UINT StepDb;                  // Step (dB)

  bool operator == (VolumeControlParamsDesc const & Other) const { return (memcmp (this, &Other, sizeof (*this)) == 0); }
  bool operator != (VolumeControlParamsDesc const & Other) const { return !operator == (Other); }

};





//
// Volume levels descriptor
// =============================================================================================
//

struct VolumeControlStateDesc {

  bool ChannelMutes [32];       // Channel mute state
  int ChannelLevels [32];       // Channel levels (dB * 65536)

};





//
// Main API classes
// =============================================================================================
//





//
// Basic driver request descriptor (header)
// =============================================================================================
//

struct DriverReq {

  UINT ApiVer;                  // API version

};





//
// Driver information request descriptor
// =============================================================================================
//

struct DriverInfoReq : public DriverReq {

};





//
// Driver information descriptor
// =============================================================================================
//

struct DriverInfo {

  UINT Flags;                   // Driver flags
  UINT VerMajor;                // Major version number
  UINT VerMinor;                // Minor version number
  UINT VerRev;                  // Revision number
  UINT VerBuild;                // Build number

  DataRangeDesc DataRange;      // Global data range

};





namespace DriverFlags {

}





__if_exists (WAVEFORMATPCMEX) {

//
// Filter information descriptor
// =============================================================================================
//

struct FilterInfo {

  UINT Flags;                   // Filter flags
  HostNotifications NotificationCode; // Host notification code
  DataFormatDesc DataFormat;    // Data format
  UINT ReservedBufferSize;      // Reserved buffer size in bytes
  UINT MinBufferFrames;         // Minimal stream buffer size in frames
  VolumeControlParamsDesc VolumeControlParams; // Volume control data
  bool HasStream;               // Has a stream
  UINT DataOverflows;           // Number of data overflows summed from all streams
  UINT DataUnderflows;          // Number of data underflows summed from all streams
  VolumeControlStateDesc VolumeState; // Current volume control state

};

} // WAVEFORMATPCMEX





namespace FilterFlags {

  UINT const StateChanged = 0x1; // Filter/stream state was changed since last notification
  UINT const WaitingForHost = 0x2; // Waiting for host confirmation
  UINT const All = 0x3;         // All flag bits

}





//
// Stream descriptor
// =============================================================================================
//

__if_exists (WAVEFORMATEXTENSIBLE) {
__if_exists (KSSTATE) {

struct StreamInfo {

  UINT Flags;                   // Stream flags
  WAVEFORMATEXTENSIBLE Fmt;     // Audio format
  KSSTATE State;                // Current state
  UINT BufferFrames;            // Buffer size in frames (0 - no buffer)
  UINT FramesBuffered;          // Amount of data buffered (frames)
  UINT64 FramesTransferred;     // Frames transferred
  UINT ClientPos;               // Client position in the buffer (frames)
  UINT DataOverflows;           // Number of data overflows
  UINT DataUnderflows;          // Number of data underflows

};

} // KSSTATE
} // WAVEFORMATPCMEX





namespace StreamFlags {

  UINT const Notifications = 1; // Notifications are used
  UINT const PacketMode = 2;    // Packet mode is used
  UINT const PositionRegister = 0x4; // Position register requested

  UINT const All = 0xf;         // All flag bits

}





//
// Two-way communication block for driver and host

struct CommonDataDesc {

  struct {

    DriverInfo DI;              // Driver information block
    FilterInfo FI;              // Filter information block
    StreamInfo SI;              // Stream information block

  } volatile Drv;               // Driver-side data

  struct {

    void * Buffer;              // Client audio buffer mapped into host's space

#ifndef _WIN64

    UINT Reserved1;             // Add unused dword to align sizes in 32/64 environments

#endif

    UINT DevicePos;             // Host (device) position in the buffer (frames)

    UINT DevicePosBytes;        // Host (device) position in the buffer (bytes)
                                // Mapped for the client as a position register

    NTSTATUS Status;            // Status code from the host

    bool Ready;                 // Host side is ready

  } volatile Host;              // Host-side data

};





//
// Special status codes
// =============================================================================================
//

namespace StatusCodes {

  static NTSTATUS const CommonDataUpdated = 0x20000001;

}





//
// Basic filter request descriptor (header)
// =============================================================================================
//

struct FilterReq : public DriverReq {

};





//
// Get common data address request descriptor
// =============================================================================================
//

struct GetCommonDataAddressReq : public FilterReq {

};





//
// Set buffer size range request descriptor
// =============================================================================================
//

struct SetMinimalBufferSizeReq : public FilterReq {

  UINT Frames;                  // Minimal frames

};





//
// Set notification event request descriptor
// =============================================================================================
//

struct SetNotificationEventReq : public FilterReq {

  HANDLE Event;                 // Event handle

};





//
// Driver notification request descriptor
// =============================================================================================
//

struct DriverNotificationReq : public FilterReq {

  NTSTATUS Code;                // Status code

};





//
// Set data range request descriptor
// =============================================================================================
//

struct SetDataFormatReq : public FilterReq {

  DataFormatDesc DataFormat;      // Data range descriptor

};





//
// Set volume control parameters request descriptor
// =============================================================================================
//

struct SetVolumeControlParamsReq : public FilterReq {

  VolumeControlParamsDesc Params; // Volume control parameters descriptor

};





//
// Set volume state request descriptor
// =============================================================================================
//

struct SetVolumeStateReq : public FilterReq {

  VolumeControlStateDesc VolumeControlState;  // Volume state descriptor

};





} // namespace





__pragma (pack (pop))
