#include "stdafx.h"





#if !UseLegacyWrappers

__pragma (comment (lib, "tagapi.lib"))

#endif





// #############################################################################################
//
// Constants
// =============================================================================================
// 
// #############################################################################################





static UINT const SamplingRateList [] = {

  5000, 8000, 11025, 22050, 44100, 48000, 88200, 96000, 192000, UINT_MAX

};

static UINT const BitsPerSampleList [] = {

  8, 16, 18, 20, 22, 24, 32, UINT_MAX

};

static UINT const NumChannelList [] = {

  1, 2, 4, 6, 8, UINT_MAX

};

static UINT const BytesPerContainerList [] = {

  0, 1, 2, 3, 4, UINT_MAX

};

static PCTSTR const ChannelConfigList [] = {

  L"def", L"0", L"1", L"3", L"4", L"3f", L"6f", L"63f", L"ff", nullptr

};

static PCTSTR const MinBufferMsList [] = {

  L"auto", L"1", L"2", L"5", L"10", L"20", L"50", nullptr

};



static UINT const FormatControls [] = {

  IDC_SamplingRate, IDC_BitsPerSample, IDC_NumChannels, IDC_BytesPerContainer,
  IDC_ChannelConfig,

  0

};

static UINT const NonRealtimeControls [] = {

  IDC_MinBufferMs, IDC_File, IDC_SelectFile, IDC_Use,

  0

};

static PCTSTR const KsStateNames [] = {L"Stop", L"Acquire", L"Pause", L"Run" };





// #############################################################################################
//
// Variables
// =============================================================================================
//
// #############################################################################################





HINSTANCE AppInst = 0;
HWND MainDlg = 0;

TagDriver * Driver;

Worker * RenderWorker;
Worker * CaptureWorker;

bool Started = false;
bool AppCloseRequest = false;



#if _Debug

  DebugContext Dbg (nullptr, 6, L"tagapp");

#endif





// #############################################################################################
//
// Worker class implementation
// =============================================================================================
//
// #############################################################################################





//
// Constructor
// =============================================================================================
//

Worker::Worker (TagDriver & Drv, bool Capt)

: Driver (Drv), Capture (Capt),

  Dlg (0),
  Host (nullptr),
  Thread (0), ThreadExitRequest (false),

  FileData (nullptr),
  FileSize (0),
  FilePos (0)

#if _Debug

  , Dbg (&::Dbg, 6, L"Wrk %c", IsCapt ()? L'c' : L'r')

#endif // _Debug

{

  *FilePath = L'\0';

}





//
// Destructor
// =============================================================================================
//

Worker::~Worker () {

  if (Dlg != 0) {

    DestroyWindow (Dlg);

  }

}





//
// Enable/disable non-realtime controls
// =============================================================================================
//

void Worker::EnableNonRealtimeControls (bool const Enable) {

  if (IsCapt ()) {

    EnableControls (Dlg, FormatControls, Enable);

  }

  EnableControls (Dlg, NonRealtimeControls, Enable);

}





//
// Read data from a PCM file
// =============================================================================================
//

HRESULT Worker::ReadDataFromFile () {

  Assert (!IsCapt ());

  Assert (FileData == nullptr && FileSize == 0 && FilePos == 0);

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    Assert (*FilePath == L'\0');

    UINT const PathLen = GetDlgItemText (Dlg, IDC_File, FilePath, NumArrayElems (FilePath));

    if (PathLen == 0) {

      Error (L"No file specified");

      break;

    }

    Res = ReadPcmFile (FilePath, Fmt, (PVOID &) (FileData), FileSize);

    if (FAILED (Res)) {

      Error (L"Cannot read PCM file %s (%s)", FilePath, ErrorText (Res));

      break;

    }

    Debug (7, L"%u PCM bytes read from %", FileSize, FilePath);

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Write data buffer to a PCM file
// =============================================================================================
//

HRESULT Worker::WriteDataToFile () {

  Assert (IsCapt ());

  Assert (FileData != nullptr && FileSize != 0 && FilePos < FileSize);

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    Assert (*FilePath != L'\0');

    Debug (7, L"Writing %u PCM bytes to %s", FilePos, FilePath);

    Res = WritePcmFile (FilePath, Fmt, FileData, FilePos);

    if (FAILED (Res)) {

      Error (L"Cannot write %u bytes of PCM data to %s (%s)", FilePath, ErrorText (Res));

      break;

    }

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Copy data from/to host buffer
// =============================================================================================
//

LocSize Worker::CopyData (void * const Addr, LocSize const Bytes) {

  Assert (Addr != nullptr && Bytes != 0);

  LocSize const Portion = min (Bytes, FileSize - FilePos);

  Assert (Portion != 0);

  if (IsCapt ()) {

    RtlCopyMemory (FileData + FilePos, Addr, Portion);

    Debug (5, L"Read %u file bytes at %u", Portion, FilePos);

  } else {

    RtlCopyMemory (Addr, FileData + FilePos, Portion);

    Debug (5, L"Written %u file bytes at %u", Portion, FilePos);

  }

#if _Debug

  LocSize const FilePosPrev = FilePos;

#endif

  FilePos += Portion;

  Assert (FilePos <= FileSize);

  if (FilePos == FileSize) {

    FilePos = 0; // Loop file buffer

  }

  Debug (5, L"File pos: %u->%u", FilePosPrev, FilePos);

  return Portion;

}





//
// Event loop
// =============================================================================================
//

HRESULT Worker::EventLoop () {

  Assert (Host != nullptr);

#if UseLegacyWrappers

  Assert (Host_IsStarted (Host));

#else

  Assert (Host->IsStarted ());

#endif

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  Assert (Dlg != 0);

  do {

    Debug (7, L"Event loop started");

#if UseLegacyWrappers

    Res = Host_EventLoop (Host, 5);

#else

    Res = Host->EventLoop (5);

#endif

    Debug (7, L"Event loop ended");

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Thread function
// =============================================================================================
//

HRESULT Worker::ThreadProc () {

  Debug (7, L"Thread started");

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  HANDLE MmcssTask = 0;

  do {

    DWORD TaskIndex = 0;

    static wchar_t MmcssClass [] = L"Pro Audio";

    MmcssTask = AvSetMmThreadCharacteristics (MmcssClass, &TaskIndex);

    if (MmcssTask) {

      Debug (7, L"Task registered in MMCSS \"%s\" class (%u)", MmcssClass, TaskIndex);

    } else {

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      Debug (9, L"Cannot register task in MMCSS \"%s\" class (%s)", MmcssClass, ErrorText (Res));

      break;

    }

    Res = EventLoop ();

  } OnlyOnce;

  Debug (7, L"Thread is exiting with %#x", Res);

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Thread function (outer)
// =============================================================================================
//

DWORD WINAPI Worker::ThreadProcOuter (void * const Arg) {

  return static_cast <DWORD> (reinterpret_cast <Worker *> (Arg)->ThreadProc ());

}





//
// Start the worker
// =============================================================================================
//

HRESULT Worker::Start () {

  Assert (Thread == 0 && !ThreadExitRequest);

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    if (IsCapt ()) {

      BOOL Success;

      UINT const SR = GetDlgItemInt (Dlg, IDC_SamplingRate, &Success, false);
      UINT const BPS = GetDlgItemInt (Dlg, IDC_BitsPerSample, &Success, false);
      UINT const NC = GetDlgItemInt (Dlg, IDC_NumChannels, &Success, false);

      UINT const BPC = GetDlgItemInt (Dlg, IDC_BytesPerContainer, &Success, false);

      wchar_t Str [16];

      GetDlgItemText (Dlg, IDC_ChannelConfig, Str, NumArrayElems (Str));

      PTSTR EndPtr = Str + 1;

      UINT const CC = (wcsicmp (Str, L"def") == 0)? UINT (-1) : wcstoul (Str, &EndPtr, 16);

      if (EndPtr == Str && *Str != L'\0') {
        
        Res = HRESULT_FROM_WIN32 (ERROR_INVALID_PARAMETER);

        Error (L"Incorrect %s channel config", GetType ());

        break;

      }

      Res = Fmt.Init (SR, BPS, NC, BPC, CC);

      if (Res != S_OK) {

        Error (L"Incorrect %s format parameters", GetType ());

        break;

      }

      UINT const PathLen = GetDlgItemText (Dlg, IDC_File, FilePath, NumArrayElems (FilePath));

      if (PathLen == 0) {

        Error (L"No %s file specified", IsCapt ()? L"output" : L"input");

        break;

      }

      Assert (FileData == nullptr && FileSize == 0);

      FileSize = Fmt.MsToBytes  (SecondsToRecord * 1000);

      FileData = new BYTE [FileSize];

      if (FileData == nullptr) {

        Res = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);

        Error (L"Cannot create capture file buffer");

        break;

      }

    } else {

      Res = ReadDataFromFile ();

      if (FAILED (Res)) {

        break;

      }

    }

    UINT const MinBufferMsIdx = static_cast <UINT> (
      SendDlgItemMessage (Dlg, IDC_MinBufferMs, CB_GETCURSEL, 0, 0)
    );

    SetDlgItemInt (Dlg, IDC_SamplingRate, Fmt.GetSR (), false);
    SetDlgItemInt (Dlg, IDC_BitsPerSample, Fmt.GetBPS (), false);
    SetDlgItemInt (Dlg, IDC_NumChannels, Fmt.GetNC (), false);
    SetDlgItemInt (Dlg, IDC_BytesPerContainer, Fmt.GetBPC (), false);

    {

      wchar_t Str [16];

      wsprintf (Str, L"%x", Fmt.GetCC ());

      SetDlgItemText (Dlg, IDC_ChannelConfig, Str);

    }

    UINT MinBufferMs = 0;

    if (MinBufferMsIdx != 0) {

      Assert (MinBufferMsIdx < NumArrayElems (MinBufferMsList));

      MinBufferMs = static_cast <UINT> (_wtoi (MinBufferMsList [MinBufferMsIdx]));

      Assert (MinBufferMs != 0);

    }

    FilePos = 0;

#if UseLegacyWrappers

    Res = Host_Create (&Host, &Driver, IsCapt (), this);

#else

    Res = Host->Create (Host, Driver, IsCapt (), *this);

#endif

    if (FAILED (Res)) {

      Error (L"Cannot create %s host (%s)", GetType (), ErrorText (Res));

      break;

    }

#if UseLegacyWrappers

    Res = Host_Start (Host, &Fmt, MinBufferMs);

#else

    Res = Host->Start (Fmt, MinBufferMs);

#endif

    if (FAILED (Res)) {

      Error (L"Cannot start %s host (%s)", GetType (), ErrorText (Res));

      break;

    }

    Thread = CreateThread (nullptr, 0, ThreadProcOuter, this, 0, nullptr);

    if (Thread == 0) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      Error (L"Cannot start %s thread (%s)", GetType (), ErrorText (Res));

      break;

    }

    Debug (7, L"Created thread");

    Res = S_OK;

  } OnlyOnce;

  if (FAILED (Res)) {

    Stop ();

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Stop the worker
// =============================================================================================
//

HRESULT Worker::Stop () {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  Debug (6, L"Stopping");

#if UseLegacyWrappers

  if (Host != nullptr && Host_IsStarted (Host)) {

    Res = Host_Stop (Host);

#else

  if (Host != nullptr && Host->IsStarted ()) {

    Res = Host->Stop ();

#endif

    if (FAILED (Res)) {

      Error (L"Cannot stop %s host (%s)", GetType (), ErrorText (Res));

    }

  }

  if (Thread != 0) {

    Assert (ThreadExitRequest == false);

    Debug (6, L"Requesting thread to exit");

    ThreadExitRequest = true;

    DWORD const WR = WaitForSingleObject (Thread, 1000);

    if (WR == WAIT_OBJECT_0) {

      Debug (6, L"Thread terminated");

    } else if (WR == WAIT_TIMEOUT) {

      Debug (9, L"Thread didn't exit, terminating it");

      TerminateThread (Thread, 9999);

    }

    Thread = 0;
    ThreadExitRequest = false;

    Res = S_OK;

  } else {

    Res = S_FALSE;

  }

  if (Host != nullptr) {

#if UseLegacyWrappers

    Host_Delete (Host);

#else

    Host->Delete ();

#endif

    Host = nullptr;

    static const UINT Ctls [] = {
      IDC_BufferSize, IDC_StreamState, IDC_FramesBuffered, IDC_FramesTransferred,
      IDC_Overflows, IDC_Underflows,
    };

    for (UINT i = 0; i < NumArrayElems (Ctls); ++i) {

      SetDlgItemText (Dlg, int (Ctls [i]), L"");

    }

  }

  if (Res == S_OK && Capture) {

    Res = WriteDataToFile ();

  }

  if (FileData != nullptr) {

    delete FileData;
    FileData = nullptr;

  }

  FileSize = 0;
  FilePos = 0;

  *FilePath = L'\0';

  Debug (7, L"%s stopped", (Res == S_OK)? L"Fully" : L"Partially");

  SetDlgItemText (Dlg, IDC_WorkerState, L"Offline");

  return Res;

}





HRESULT Worker::SelectInputFile () {

  wchar_t TempPath [MAX_PATH];

  *TempPath = L'\0';

  OPENFILENAME OF;

  ZeroObject (OF);

  OF.lStructSize = sizeof (OF);
  OF.hwndOwner = Dlg;
  OF.hInstance = AppInst;
  OF.lpstrInitialDir = nullptr;
  OF.lpstrFilter = L"Wave files (*.wav)\0*.wav\0";
  OF.lpstrFile = TempPath;
  OF.nMaxFile = NumArrayElems (TempPath);
  OF.lpstrTitle = L"Select a file to load";
  OF.lpstrDefExt = L"wav";
  OF.Flags = OFN_DONTADDTORECENT;

  HRESULT Res;

  if (GetOpenFileName (&OF)) {

    Debug (7, L"File load path: \"%s\"", TempPath);

    SetDlgItemText (Dlg, IDC_File, TempPath);

    Res = S_OK;

  } else {

    Debug (9, L"Input file selection failed (%#x)", CommDlgExtendedError ());

    Res = E_FAIL;

  }

  return Res;

}





HRESULT Worker::SelectOutputFile () {

  wchar_t TempPath [MAX_PATH];

  *TempPath = L'\0';

  OPENFILENAME OF;

  ZeroObject (OF);

  OF.lStructSize = sizeof (OF);
  OF.hwndOwner = Dlg;
  OF.hInstance = AppInst;
  OF.lpstrInitialDir = nullptr;
  OF.lpstrFilter = L"Wave files (*.wav)\0*.wav\0";
  OF.lpstrFile = TempPath;
  OF.nMaxFile = NumArrayElems (TempPath);
  OF.lpstrTitle = L"Select a file to save data to";
  OF.lpstrDefExt = L"cfg";
  OF.Flags = OFN_DONTADDTORECENT | OFN_OVERWRITEPROMPT;

  HRESULT Res;

  if (GetSaveFileName (&OF)) {

    Debug (8, L"File save path: \"%s\"", TempPath);

    SetDlgItemText (Dlg, IDC_File, TempPath);

    Res = S_OK;

  } else {

    Debug (9, L"Output file selection failed (%x)", CommDlgExtendedError ());

    Res = E_FAIL;

  }

  return Res;

}





//
// Pipe dialog function
// =============================================================================================
//

INT_PTR Worker::HostDialogProc (HWND const Dlg, UINT const Msg, WPARAM const WP, LPARAM const LP) {

  Unreferenced (LP);

  INT_PTR Res = 0;

  switch (Msg) {


    case WM_INITDIALOG: {

      SetDlg (Dlg);

      Debug (6, L"PipeDlg %s - init, HWND %#x", IsCapt ()? L"capture" : L"render", Dlg);

      SetDlgItemText (
        Dlg, IDC_Caption, IsCapt ()? L"Capture pipe (render KS pin)" : L"Render pipe (capture KS pin)"
      );

      Fmt.Init (SamplingRateDef, BitsPerSampleDef, NumChannelsDef);

      if (IsCapt ()) {

        SetDlgItemText (Dlg, IDC_FileType, L"Save to");
        SetDlgItemText (Dlg, IDC_File, FileToRecordDef);

      } else {

        EnableControls (Dlg, FormatControls, false);

        SetDlgItemText (Dlg, IDC_FileType, L"Load from");
        SetDlgItemText (Dlg, IDC_File, FileToPlayDef);

      }

      FillComboBox (Dlg, IDC_SamplingRate, SamplingRateList);
      FillComboBox (Dlg, IDC_BitsPerSample, BitsPerSampleList);
      FillComboBox (Dlg, IDC_NumChannels, NumChannelList);
      FillComboBox (Dlg, IDC_BytesPerContainer, BytesPerContainerList);
      FillComboBox (Dlg, IDC_ChannelConfig, ChannelConfigList);

      SetDlgItemInt (Dlg, IDC_SamplingRate, Fmt.GetSR (), false);
      SetDlgItemInt (Dlg, IDC_BitsPerSample, Fmt.GetBPS (), false);
      SetDlgItemInt (Dlg, IDC_NumChannels, Fmt.GetNC (), false);
      SetDlgItemInt (Dlg, IDC_BytesPerContainer, 0, false);

#if 1

      SetDlgItemText (Dlg, IDC_ChannelConfig, L"def");

#else

      wchar_t Str [16];

      SetDlgItemText (
        Dlg, IDC_ChannelConfig, _itow (static_cast <int> (Fmt.GetCC ()), Str, 16)
      );

#endif

      FillComboBox (Dlg, IDC_MinBufferMs, MinBufferMsList);
      SendDlgItemMessage (Dlg, IDC_MinBufferMs, CB_SETCURSEL, 0, 0);

      SendDlgItemMessage (Dlg, IDC_Use, BM_SETCHECK, BST_CHECKED, 0);

      ShowWindow (Dlg, SW_SHOW);

      break;

    }



    case WM_COMMAND: {

      UINT const CtrlId = LOWORD (WP);
      UINT const CtrlMsg = HIWORD (WP);

      Unreferenced (CtrlMsg);

      if (CtrlId == IDC_SelectFile) {

        if (IsCapt ()) {

          SelectOutputFile ();

        } else {

          SelectInputFile ();

        }

      }

      break;

    }



    case WM_TIMER: {

      Debug (5, L"Timer");

      if (
        Host != nullptr

#if UseLegacyWrappers

        && Host_IsStarted (Host)

#else

        && Host->IsStarted ()

#endif

      ) {

        CommonDataDesc const & CD = (

#if UseLegacyWrappers

          *Host_GetCommonData (Host)

#else

          Host->GetCommonData ()

#endif

        );

        StreamInfo const volatile & SI = CD.Drv.SI;

        SetDlgItemText (Dlg, IDC_WorkerState, CD.Host.Ready? L"Ready" : L"Offline");
    
        SetDlgItemInt (Dlg, IDC_FramesBuffered, SI.FramesBuffered, false);
        SetDlgItemInt (Dlg, IDC_Overflows, SI.DataOverflows, false);
        SetDlgItemInt (Dlg, IDC_Underflows, SI.DataUnderflows, false);
        SetDlgItemInt (Dlg, IDC_BufferSize, SI.BufferFrames, false);

        wchar_t Text [32];

        wsprintf (Text, L"%I64u", CD.Drv.SI.FramesTransferred);
        SetDlgItemText (Dlg, IDC_FramesTransferred, Text);

        Assert (SI.State <= NumArrayElems (KsStateNames));
        SetDlgItemText (Dlg, IDC_StreamState, KsStateNames [SI.State]);

      }

      break;

    }

    default: {

      //Debug (6, L"PipeDlg - msg %#x", Msg);

      break;

    }

  }

  return Res;

}





//
// Pipe dialog function (outer)
// =============================================================================================
//

INT_PTR WINAPI HostDialogProcOuter (HWND const Dlg, UINT const Msg, WPARAM const WP, LPARAM const LP) {

  Worker * Wrk = nullptr;

  if (Msg == WM_INITDIALOG) {

    SetWindowLongPtr (Dlg, DWLP_USER, LP);

  }

  Wrk = reinterpret_cast <Worker *> (GetWindowLongPtr (Dlg, DWLP_USER));

  return (Wrk != nullptr)? Wrk->HostDialogProc (Dlg, Msg, WP, LP) : FALSE;

}





HRESULT StopAll () {

  KillTimer (MainDlg, 1);

  CaptureWorker->Stop ();
  RenderWorker->Stop ();

  RenderWorker->EnableNonRealtimeControls (true);
  CaptureWorker->EnableNonRealtimeControls (true);
  SetDlgItemText (MainDlg, IDC_StartStop, L"Start");
  UpdateWindow (MainDlg);

#if UseLegacyWrappers

  Driver_Close (Driver);

#else

  Driver->Close ();

#endif

  Started = false;

  return S_OK;

}





HRESULT StartAll () {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  Assert (!Started);

  EnableWindow (GetDlgItem (MainDlg, IDC_StartStop), false);
  RenderWorker->EnableNonRealtimeControls (false);
  CaptureWorker->EnableNonRealtimeControls (false);
  UpdateWindow (MainDlg);

  do {

    SetTimer (MainDlg, 1, 200, nullptr);

#if UseLegacyWrappers

    Res = Driver_Open (Driver);

#else

    Res = Driver->Open ();

#endif

    if (Res != S_OK) {

      if (Res == HRESULT_FROM_WIN32 (ERROR_NOT_FOUND)) {

        Error (L"No audio driver interface found");

      } else {

        Error (L"Cannot open audio driver (%s)", ErrorText (Res));

      }

      break;

    }

    if (SendDlgItemMessage (RenderWorker->GetDlg (), IDC_Use, BM_GETCHECK, 0, 0) == BST_CHECKED) {

      Res = RenderWorker->Start ();

      if (FAILED (Res)) {

        break;

      }

    }

    if (SendDlgItemMessage (CaptureWorker->GetDlg (), IDC_Use, BM_GETCHECK, 0, 0) == BST_CHECKED) {

      Res = CaptureWorker->Start ();

      if (FAILED (Res)) {

        break;

      }

    }



    Started = true;

    Debug (7, L"Workers started");

    SetDlgItemText (MainDlg, IDC_StartStop, L"Stop");

    Res = S_OK;

  } OnlyOnce;

  if (FAILED (Res)) {

    StopAll ();

  }

  EnableWindow (GetDlgItem (MainDlg, IDC_StartStop), true);

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Main dialog function
// =============================================================================================
//

INT_PTR WINAPI MainDialogProc (HWND const Dlg, UINT const Msg, WPARAM const WP, LPARAM const LP) {

  INT_PTR Res = 0;

  switch (Msg) {

    case WM_INITDIALOG: {

      Unreferenced (LP);

      Debug (6, L"MainDlg - init, HWND %#x", Dlg);

      MainDlg = Dlg;

      HWND const RenderDlg = CreateDialogParam (
        AppInst, MAKEINTRESOURCE (IDD_Pipe), Dlg,
        HostDialogProcOuter, reinterpret_cast <LPARAM> (RenderWorker)
      );

      Assert (RenderDlg != 0);

      HWND const RenderCtlWin = GetDlgItem (Dlg, IDC_RenderDialog);

      RECT R;

      GetWindowRect (RenderCtlWin, &R);
      ScreenToClient (Dlg, PPOINT (&R));
      SetWindowPos (RenderDlg, RenderCtlWin, R.left, R.top, 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW);

      HWND const CaptureDlg = CreateDialogParam (
        AppInst, MAKEINTRESOURCE (IDD_Pipe), Dlg,
        HostDialogProcOuter, reinterpret_cast <LPARAM> (CaptureWorker)
      );

      Assert (CaptureDlg != 0);

      HWND const CaptureCtlWin = GetDlgItem (Dlg, IDC_CaptureDialog);

      GetWindowRect (CaptureCtlWin, &R);
      ScreenToClient (Dlg, PPOINT (&R));
      SetWindowPos (CaptureDlg, CaptureCtlWin, R.left, R.top, 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW);

      //UINT const PipeDlgWidth = R.right - R.left + 1;
      //UINT const PipeDlgHeight = R.bottom - R.top + 1;

      ShowWindow (Dlg, SW_SHOW);

      break;

    }

    case WM_COMMAND: {

      UINT const CtrlId = LOWORD (WP);
      UINT const CtrlMsg = HIWORD (WP);

      Unreferenced (CtrlMsg);

      if (CtrlId == IDC_StartStop) {

        if (!Started) {

          HRESULT const LocRes = StartAll ();

          if (FAILED (LocRes)) {

            StopAll ();

            break;

          }

          Assert (Started);

        } else { // Start/stop

          Debug (6, L"MainDlg - stop");

          StopAll ();

        }

      } else if (CtrlId == IDCANCEL) {

        Debug (6, L"MainDlg - close/exit");

        StopAll ();

        AppCloseRequest = true;

      } else {

        HWND const W = reinterpret_cast <HWND> (LP);

        if (W == RenderWorker->GetDlg () || W == CaptureWorker->GetDlg ()) {

          SendMessage (W, WM_COMMAND, WP, LP);

        }

      }
      
      break;

    } // WM_COMMAND


    case WM_TIMER: {

      Debug (5, L"Timer");

      SendMessage (RenderWorker->GetDlg (), WM_TIMER, 0, 0);
      SendMessage (CaptureWorker->GetDlg (), WM_TIMER, 0, 0);

      break;

    }


    case WM_CHAR: {

      Debug (7, L"WM_CHAR for %x", LP);

      break;

    }

  } // message switch

  return Res;

}





//
// Main function
// =============================================================================================
//

HRESULT Main () {

  HRESULT FinalRes = E_I_PROCESSING_NOT_FINISHED;

  HWND Dlg = 0;

  do {

    HRESULT HR;

    struct __declspec (uuid (ProductGuidStr)) ProductGuidStruct;

    GUID const & ProductGuid = __uuidof (ProductGuidStruct);

    Assert (Driver == nullptr);

#if UseLegacyWrappers

    HR = Driver_Create (&Driver, &ProductGuid);

#else

    // Static methods can be called even via invalid object pointer

    HR = Driver->Create (Driver, ProductGuid);

#endif

    if (FAILED (HR)) {

      Error (L"Cannot create driver object (%s)", ErrorText (HR));

      FinalRes = HR;
      break;

    }

    Assert (CaptureWorker == nullptr);

    CaptureWorker = new Worker (*Driver, true);

    if (CaptureWorker == nullptr) {

      Error (L"No memory for capture worker object");

      FinalRes = E_OUTOFMEMORY;
      break;

    }

    Assert (RenderWorker == nullptr);

    RenderWorker = new Worker (*Driver, false);

    if (RenderWorker == nullptr) {

      Error (L"No memory for render worker object");

      FinalRes = E_OUTOFMEMORY;
      break;

    }

    Assert (MainDlg == 0);

    Dlg = CreateDialog (AppInst, MAKEINTRESOURCE (IDD_Main), HWND_DESKTOP, MainDialogProc);

    if (Dlg == 0) {

      MessageBox (HWND_DESKTOP, L"Cannot create dialog", L"testapp", MB_ICONERROR | MB_OK);

      FinalRes = E_FAIL;
      break;

    }

    Assert (MainDlg == Dlg);

    Debug (7, L"Starting window message loop");

    MSG Msg;

    while (GetMessage (&Msg, 0, 0, 0) > 0 && !AppCloseRequest) {

      //Debug (7, L"Msg %#x", Msg.message);

      if (Msg.message == WM_CHAR) {

        Debug (7, L"WM_CHAR for %x", Msg.hwnd);

      }

      if (
        !IsDialogMessage (Dlg, &Msg)
        //&& RenderWorker->GetDlg () != 0 && !IsDialogMessage (RenderWorker->GetDlg (), &Msg)
        //&& CaptureWorker->GetDlg () != 0 && !IsDialogMessage (CaptureWorker->GetDlg (), &Msg)
      ) {

        TranslateMessage (&Msg);
        DispatchMessage (&Msg);

      }

    }

    FinalRes = S_OK;

  } OnlyOnce;

  if (Dlg != 0) {

    DestroyWindow (Dlg);
    Dlg = 0;

  }

  if (RenderWorker != nullptr) {

    delete RenderWorker;
    RenderWorker = nullptr;

  }

  if (CaptureWorker != nullptr) {

    delete CaptureWorker;
    CaptureWorker = nullptr;

  }

  if (Driver != nullptr) {

#if UseLegacyWrappers

    Driver_Delete (Driver);

#else

    Driver->Delete ();

#endif

    Driver = nullptr;

  }

  Assert (FinalRes != E_I_PROCESSING_NOT_FINISHED);

  return FinalRes;

}





//
// Entry function
// =============================================================================================
//

int WINAPI WinMain (HINSTANCE const Inst, HINSTANCE const PrevInst, PSTR const Cmd, int const Show) {

  (void) (PrevInst && Cmd && Show);

  AppInst = Inst;

#if _Debug

  Dbg.SetPrintLevel (6);

#endif

  Debug (7, L"Started");

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    Res = Main ();

  } OnlyOnce;


  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return int (Res);

}
