#include "stdafx.h"





// #############################################################################################
//
// Functions
// =============================================================================================
//
// #############################################################################################





void FillComboBox (HWND const Dlg, UINT const Id, UINT const Values []) {

  for (UINT const * p = Values; *p != UINT_MAX; ++p) {

    wchar_t Text [16];
  
    wsprintf (Text, L"%u", *p);

    SendDlgItemMessage (Dlg, static_cast <int> (Id), CB_ADDSTRING, 0, reinterpret_cast <LPARAM> (Text));

  }

}





void FillComboBox (HWND const Dlg, UINT const Id, PCTSTR const Values []) {

  for (PCTSTR const * p = Values; *p != nullptr; ++p) {

    SendDlgItemMessage (Dlg, static_cast <int> (Id), CB_ADDSTRING, 0, reinterpret_cast <LPARAM> (*p));

  }

}





void EnableControls (HWND const Dlg, UINT const Ids [], bool const Enable) {

  for (UINT const * p = Ids; *p != 0; ++p) {

    EnableWindow (GetDlgItem (Dlg, static_cast <int> (*p)), Enable);

  }

}





//
// Read a PCM file into memory
// =============================================================================================
//

HRESULT ReadPcmFile (PCTSTR const FilePath, WaveFormat & Fmt, PVOID & ForData, LocSize & ForSize) {

  Debug (6, L"Reading %s", FilePath);

  HMMIO File = 0;
  void * Buffer = nullptr;

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    MMIOINFO MI;

    memset (&MI, 0, sizeof (MI));

    File = mmioOpen (
      const_cast <PTSTR> (FilePath),
      &MI,
      MMIO_READ | MMIO_ALLOCBUF | MMIO_DENYNONE
    );

    if (!File) {

      Res = HRESULT_FROM_WIN32 (ERROR_FILE_NOT_FOUND);

      Debug (9, L"Cannot open file %s (%u)", FilePath, MI.wErrorRet);

      break;

    }


    MMCKINFO RiffChunk;

    RiffChunk.fccType = mmioFOURCC (L'W', L'A', L'V', L'E');

    MMRESULT MmRes;

    MmRes = mmioDescend (File, &RiffChunk, 0, MMIO_FINDRIFF);

    if (MmRes != MMSYSERR_NOERROR) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"File contain no RIFF/WAVE structure (%u)", MmRes);

      break;

    }


    MMCKINFO FmtChunk;

    FmtChunk.ckid = mmioFOURCC (L'f', L'm', L't', L' ');

    MmRes = mmioDescend (File, &FmtChunk, &RiffChunk, MMIO_FINDCHUNK);

    if (MmRes != MMSYSERR_NOERROR) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"File contain no format chunk (%u)", MmRes);

      break;

    }


    WAVEFORMATEXTENSIBLE Wfx;

    Debug (7, L"Format chunk size: %u", FmtChunk.cksize);

    if (FmtChunk.cksize > sizeof (Wfx)) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"Format descriptor is too big (%u)", FmtChunk.cksize);

      break;

    }


    LONG Len = mmioRead (File, reinterpret_cast <PCHAR> (&Wfx), LONG (FmtChunk.cksize));

    if (Len == -1 || static_cast <DWORD> (Len) != FmtChunk.cksize) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"Invalid format descriptor");

      break;

    }

    Res = Fmt.FromWfx (Wfx.Format);

    if (FAILED (Res)) {

      Debug (9, L"Unsupported file format");

      break;

    }

    Debug (7, L"File format: %s", WfxText (Fmt));

    MMCKINFO DataChunk;

    DataChunk.ckid = mmioFOURCC (L'd', L'a', L't', L'a');

    MmRes = mmioDescend (File, &DataChunk, &RiffChunk, MMIO_FINDCHUNK);

    if (MmRes != MMSYSERR_NOERROR) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"Cannot descend into data chunk (%u)", MmRes);

      break;

    }


    Debug (7, L"Data chunk size: %u", DataChunk.cksize);

    Buffer = new BYTE [DataChunk.cksize];

    if (!Buffer) {

      Res = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);

      Debug (9, L"Cannot allocate %u bytes for PCM file data", DataChunk.cksize);

      break;

    }


    Len = mmioRead (File, reinterpret_cast <PCHAR> (Buffer), LONG (DataChunk.cksize));

    if (Len == -1 || static_cast <DWORD> (Len) != DataChunk.cksize) {

      Res = HRESULT_FROM_WIN32 (ERROR_READ_FAULT);

      Debug (9, L"Error reading PCM data");

      break;

    }


    ForData = Buffer;
    ForSize = DataChunk.cksize;

    Res = S_OK;

  } OnlyOnce;


  if (FAILED (Res) && Buffer != nullptr) {

    delete Buffer;

    Buffer = nullptr;

  }

  if (File != 0) {

    mmioClose (File, 0);

  }


  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





//
// Write data to a PCM file
// =============================================================================================
//

HRESULT WritePcmFile (PCTSTR const FilePath, WaveFormat const & Fmt, PCVOID const Data, LocSize const Size) {

  Debug (
    6,
    L"Writing %u bytes (%u frames) to %s",
    Size, Fmt.BytesToFrames (static_cast <UINT> (Size)), FilePath
  );

  HMMIO File = 0;

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    File = mmioOpen (
      PTSTR (FilePath),
      nullptr,
      MMIO_CREATE | MMIO_READWRITE /*| MMIO_ALLOCBUF*/ | MMIO_DENYWRITE
    );

    if (!File) {

      Res = HRESULT_FROM_WIN32 (ERROR_CANNOT_MAKE);

      Debug (9, L"Cannot create %s", FilePath);

      break;

    }

    MMCKINFO RiffChunk, FmtChunk, DataChunk;

    ZeroMemory (&RiffChunk, sizeof (RiffChunk));
    ZeroMemory (&FmtChunk, sizeof (FmtChunk));
    ZeroMemory (&DataChunk, sizeof (DataChunk));

    RiffChunk.fccType = mmioFOURCC (L'W', L'A', L'V', L'E');

    MMRESULT MmRes;

    MmRes = mmioCreateChunk (File, &RiffChunk, MMIO_CREATERIFF);

    if (MmRes != MMSYSERR_NOERROR) {

      Res = HRESULT_FROM_WIN32 (ERROR_CANNOT_MAKE);

      Debug (9, L"Cannot create RIFF chunk (%u)", MmRes);

      break;

    }

    WAVEFORMATEXTENSIBLE Wfx;

    Fmt.ToWfx (Wfx);

    FmtChunk.ckid = mmioFOURCC (L'f', L'm', L't', L' ');
    FmtChunk.cksize = sizeof (Wfx.Format) + Wfx.Format.cbSize;;

    MmRes = mmioCreateChunk (File, &FmtChunk, 0);

    if (MmRes != MMSYSERR_NOERROR) {

      Res = HRESULT_FROM_WIN32 (ERROR_CANNOT_MAKE);

      Debug (9, L"Cannot create fmt chunk (%u)", MmRes);

      break;

    }

    LONG Written;

    Written = mmioWrite (File, reinterpret_cast <PSTR> (&Wfx), static_cast <LONG> (FmtChunk.cksize));

    if (DWORD (Written) != FmtChunk.cksize) {

      Res = HRESULT_FROM_WIN32 (ERROR_WRITE_FAULT);

      Debug (9, L"Cannot write format data");

      break;

    }

    MmRes = mmioAscend (File, &FmtChunk, 0);

    if (MmRes != MMSYSERR_NOERROR) {

      Res = HRESULT_FROM_WIN32 (ERROR_WRITE_FAULT);

      Debug (9, L"Cannot ascend out from fmt chunk (%u)", MmRes);

      break;

    }

    DataChunk.ckid = mmioFOURCC (L'd', L'a', L't', L'a');
    DataChunk.cksize = static_cast <DWORD> (Size);

    MmRes = mmioCreateChunk (File, &DataChunk, 0);

    if (MmRes != MMSYSERR_NOERROR) {

      Res = HRESULT_FROM_WIN32 (ERROR_WRITE_FAULT);

      Debug (9, L"Cannot create data chunk (%u)", MmRes);

      break;

    }

    Written = mmioWrite (File, static_cast <PSTR> (const_cast <PVOID> (Data)), static_cast <LONG> (Size));

    if (static_cast <ULONG> (Written) != Size) {

      Res = HRESULT_FROM_WIN32 (ERROR_WRITE_FAULT);

      Debug (9, L"Cannot write wave data to file");

      break;

    }

    MmRes = mmioAscend (File, &DataChunk, 0);

    if (MmRes != MMSYSERR_NOERROR) {

      Res = HRESULT_FROM_WIN32 (ERROR_WRITE_FAULT);

      Debug (9, L"Cannot ascend out from data chunk (%u)", MmRes);

      break;

    }

    MmRes = mmioAscend (File, &RiffChunk, 0);

    if (MmRes != MMSYSERR_NOERROR) {

      Res = HRESULT_FROM_WIN32 (ERROR_WRITE_FAULT);

      Debug (9, L"Cannot ascend out from RIFF chunk (%u)", MmRes);

      break;

    }

    Res = S_OK;

  } OnlyOnce;

  if (File != 0) {

    mmioClose (File, 0);

    File = 0;

    if (FAILED (Res)) {

      DeleteFile (FilePath);

    }

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}
