#include "common.h"





//
// Check if wchar_t is blank
// =============================================================================================
//

bool IsBlank (wchar_t c) {

  return (c == L' ' || c == L'\t');

}





//
// Find blank wchar_t in the string
// =============================================================================================
//

PTSTR FindBlank (PTSTR p) {

  while (*p && !IsBlank (*p)) p++;

  return p;

}





//
// Skip blank chars in the string
// =============================================================================================
//

PCTSTR SkipBlank (PCTSTR p) {

  while (IsBlank (*p)) p++;

  return p;

}





//
// Compare two strings with length limitation
// =============================================================================================
//

int _cdecl strncmpi (PCTSTR S1, PCTSTR S2, size_t Len) {

  return (
    CompareString (
      LOCALE_USER_DEFAULT,
      NORM_IGNORECASE,
      S1,
      int (Len),
      S2,
      int (Len)
    )
    - 2
  );

}





//
// Retrieve a quoted string
// =============================================================================================
//
// Retrieves a string, possibly enclosed in double quotemarks
//
// Function returns a pointer beyond right quotemark or at space delimeter, or nullptr in case of error.
//
// If Buffer is zero, a substring is skipped.
//

PCTSTR GetQuotedString (PCTSTR p, PTSTR Buffer) {

  PCTSTR Next = nullptr;

  do {

    p = SkipBlank (p);

    if (!*p) {

      Debug (9, L"No substring");

      break;

    }

    wchar_t Delim;

    if (*p == L'\"') {
  
      Delim = *p++;

    } else {

      Delim = L' ';

    }

    Next = wcschr (p, Delim);

    if (!Next) {
    
      if (Delim == L'\"') {

        Debug (9, L"No right quotemark");

        break;

      }

      Next = p + wcslen (p);

    }

    if (Buffer != nullptr) {

      size_t const Len = size_t (Next - p);

      wcsncpy (Buffer, p, Len);

      Buffer [Len] = L'\0';

    }

    if (*Next == L'\"') {

      Next++;

    }

    Next = SkipBlank (Next);

  } OnlyOnce;

  return Next;

}





bool DelFile (PCTSTR const Path) {

  bool Success = false;

  do {

    DWORD const Attr = GetFileAttributes (Path);

    if (Attr == INVALID_FILE_ATTRIBUTES) {

      if (GetLastError () == ERROR_FILE_NOT_FOUND || GetLastError () == ERROR_PATH_NOT_FOUND) {

        Debug (7, L"No file \"%s\"", Path);

        break;

      } else {

        Debug (9, L"Cannot get attributes of \"%s\" (%s)", Path, LastErrorText ());

      }

    } else if (TestBits (Attr, FILE_ATTRIBUTE_READONLY)) {

      Debug (7, L"Clearing RO attribute of \"%s\"", Path);

      if (!SetFileAttributes (Path, Attr & ~FILE_ATTRIBUTE_READONLY)) {

        Debug (9, L"Cannot clear RO attribute of \"%s\" (%s)", Path, LastErrorText ());

      }

    }

    Debug (7, L"Deleting \"%s\"", Path);

    if (!DeleteFile (Path)) {
      
      if (GetLastError () != ERROR_FILE_NOT_FOUND && GetLastError () != ERROR_PATH_NOT_FOUND) {

        Warning (L"Cannot delete \"%s\" (%s)", Path, LastErrorText ());

        break;

      } else {

        Debug (7, L"No file \"%s\"", Path);

        break;

      }

    }

    Success = true;

  } OnlyOnce;

  return Success;

}





bool GetPreferredDevice (bool const Out, PTSTR const Buffer) {

#if _Debug

  PCTSTR const Type = Out? L"Out" : L"In";

#endif

  DWORD Id, Flags;

  MMRESULT Res = Out? (
    waveOutMessage (
      HWAVEOUT (INT_PTR (WAVE_MAPPER)),
      DRVM_MAPPER_PREFERRED_GET,
      DWORD_PTR (&Id), DWORD_PTR (&Flags)
    )
  ) : (
    waveInMessage (
      HWAVEIN (INT_PTR (WAVE_MAPPER)),
      DRVM_MAPPER_PREFERRED_GET,
      DWORD_PTR (&Id), DWORD_PTR (&Flags)
    )
  );

  if (Res == MMSYSERR_NOERROR) {
  
    Debug (7, L"Preferred Wave %s device ID is %d", Type, Id);

  } else {

    Debug (9, L"Cannot get preferred Wave %s device ID (%u)", Type, Res);

  }

  if (Res != MMSYSERR_NOERROR || Id == -1) {

    Debug (7, L"Assuming preferred Wave %s device ID is 0", Type);

    Id = 0;

  }

  WAVEOUTCAPS Caps;

  Res = Out? (
    waveOutGetDevCaps (Id, &Caps, sizeof (Caps))
  ) : (
    waveInGetDevCaps (Id, (WAVEINCAPS *) (&Caps), sizeof (WAVEINCAPS))
  );

  if (Res == MMSYSERR_NOERROR) {

    wcscpy (Buffer, Caps.szPname);

    Debug (7, L"Wave %s device name is \"%s\"", Type, Buffer);

    return true;

  } else {

    Debug (9, L"Cannot get Wave %s caps for %u (%u)", Type, Id, Res);

    return false;

  }

}





bool SetPreferredDevice (bool const Out, PCTSTR const Name) {

  wchar_t const SoundMapper [] = L"Software\\Microsoft\\Multimedia\\Sound Mapper";

  bool Success = false;

  HKEY Key = 0;

  do {

    LONG Res = RegCreateKey (HKEY_CURRENT_USER, SoundMapper, &Key);

    if (Res != ERROR_SUCCESS && Res != ERROR_ALREADY_EXISTS) {

      Debug (9, L"Cannot open/create Sound Mapper key (%s)", ErrorText (Res));

      break;

    }

    PCTSTR const ValueName = Out? L"Playback" : L"Record";

    wchar_t ValueData [128];

    DWORD ValueSize = sizeof (ValueData);

    ZeroMemory (ValueData, ValueSize);

    Res = RegQueryValueEx (Key, ValueName, 0, 0, PBYTE (ValueData), &ValueSize);

    if (Res != ERROR_FILE_NOT_FOUND) {

      if (Res == ERROR_SUCCESS) {

        Debug (
          9,
          L"Value \"%s\" already exists - \"%s\"",
          ValueName, ValueData
        );

      } else {

        Debug (
          9,
          L"Value \"%s\" returns unexpected error (%s)",
          ValueName, ErrorText (Res)
        );

      }

      break;

    }

    Res = RegSetValueEx (Key, ValueName, 0, REG_SZ, PCBYTE (Name), StrLenToSz (wcslen (Name)));

    if (Res != ERROR_SUCCESS) {

      Debug (9, L"Cannot set value \"%s\" (%s)", ValueName, ErrorText (Res));

      break;

    }

    Debug (7, L"Value \"%s\" set to \"%s\"", ValueName, Name);

    Success = true;

  } OnlyOnce;

  if (Key != 0) {

    RegCloseKey (Key);

    Key = 0;

  }

  return Success;

}





void TranslateSetupApiError (void) {

  DWORD const OldCode = GetLastError ();

  DWORD const Mask = APPLICATION_ERROR_MASK | ERROR_SEVERITY_ERROR;

  if (TestBits (OldCode, Mask) == Mask) {

    DWORD const NewCode = DWORD (HRESULT_FROM_SETUPAPI (OldCode));

    if (NewCode != OldCode) {

  	  SetLastError (NewCode);

      Debug (
        6,
        L"Translated SetupAPI error code %X to %X",
        OldCode, NewCode
      );

    }

  }

}





void FlushVolumes (void) {

  Debug (7, L"Flushing volume buffers");

  wchar_t VolName [MAX_PATH];

  HANDLE const VolEnum = FindFirstVolume (VolName, NumArrayElems (VolName));

  if (VolEnum != INVALID_HANDLE_VALUE) {

    do {

      UINT const Type = GetDriveType (VolName);

      Debug (5, L"Vol: %s, type: %X", VolName, Type);

      if (Type == DRIVE_FIXED || Type == DRIVE_REMOVABLE || Type == DRIVE_REMOTE) {

        size_t const Len = size_t (wcslen (VolName));

        if (Len != 0 && VolName [Len - 1] == L'\\') {

          VolName [Len - 1] = L'\0';

        }

        HANDLE VolHandle = 0;

        do {

          VolHandle = CreateFile (
            VolName,
            GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE,
            nullptr,
            OPEN_EXISTING,
            0,
            0
          );

          if (VolHandle == INVALID_HANDLE_VALUE) {

            if (GetLastError () != ERROR_FILE_NOT_FOUND) {

              Debug (9, L"Cannot open %s (%s)", VolName, LastErrorText ());

            }

            VolHandle = 0;

            break;

          }

          Debug (6, L"Flushing %s", VolName);

          if (FlushFileBuffers (VolHandle)) {
            
            Debug (6, L"Flushed %s", VolName);

          } else {

            if (Type == DRIVE_FIXED || GetLastError () != ERROR_INVALID_FUNCTION) {

              Debug (9, L"Cannot flush %s (%s)", VolName, LastErrorText ());

            }

            break;

          }

        } OnlyOnce;

        if (VolHandle != 0) {

          CloseHandle (VolHandle);

          VolHandle = 0;

        }

      }

    } while (FindNextVolume (VolEnum, VolName, NumArrayElems (VolName)));

    if (GetLastError () != ERROR_NO_MORE_FILES) {

      Debug (9, L"Cannot find next volume (%s)", LastErrorText ());

    }

    Assert (FindVolumeClose (VolEnum));

  }

}





//
// Flush known registry keys/hives
// =============================================================================================
//

void FlushRegistry () {

  Debug (7, L"Flushing registry buffers");

  static const HKEY Keys [] = {
    HKEY_CLASSES_ROOT,
    HKEY_CURRENT_CONFIG,
    HKEY_CURRENT_USER,
    HKEY_LOCAL_MACHINE,
    HKEY_PERFORMANCE_DATA,
    HKEY_USERS,
  };

  for (UINT i = 0; i < NumArrayElems (Keys); ++i) {

    LONG Res = RegFlushKey (Keys [i]);

    if (Res == NO_ERROR) {

      Debug (6, L"Flushed key %x", Keys [i]);

    } else {

      Debug (9, L"Cannot flush key %x (%s)", Keys [i], ErrorText (Res));

    }

  }

}





//
// Flush all temporary data
// =============================================================================================
//

void FlushStorage () {

  FlushRegistry ();
  FlushVolumes ();

}





HDEVINFO CreateDL (GUID const * const ClassGuid, DWORD const Flags) {

  HDEVINFO DL;

  do {

    DL = SetupDiGetClassDevs (ClassGuid, 0, 0, Flags);

    if (DL == INVALID_HANDLE_VALUE) {

      DL = 0;

      TranslateSetupApiError ();

      Error (L"Cannot create device info list (%s)", LastErrorText ());

      break;

    }

  } OnlyOnce;

  return DL;

}





void DestroyDL (HDEVINFO & DL) {

  if (DL != 0) {

    SetupDiDestroyDeviceInfoList (DL);

    DL = 0;

  }

}





//
// Get registry value
// =============================================================================================
//

bool GetRegistrySzValue (HKEY const Key, PCTSTR const Name, PTSTR const Value, DWORD const Size) {

  bool Success;

  DWORD Type;
  DWORD Sz = Size;

  LONG const Res = RegQueryValueEx (
    Key, Name, 0, &Type, PBYTE (Value), &Sz
  );

  if (Res == NO_ERROR && Type == REG_SZ) {

    Success = true;

  } else {

    if (Res != ERROR_FILE_NOT_FOUND) {

      Debug (9, L"Cannot get %s (%s)", Name, ErrorText (Res));

    }

    Success = false;

  }

  return Success;

}





bool DoesDeviceIdMatch (HDEVINFO const DL, SP_DEVINFO_DATA const & DI, PCTSTR const Id, bool const ByInstanceId) {

  bool Success = false;

  PTSTR Buffer = nullptr;

  do {

    UINT BufferSize = 0;
    bool CompatibleIds = false;

    for (;;) {

      PCTSTR const IdType = ByInstanceId? L"instance" : CompatibleIds? L"compatible" : L"hardware";

      bool Res;
      DWORD DataType;
      DWORD DataSize;

      if (ByInstanceId) {
        
        Assert (BufferSize % 2 == 0); // SetupDiGetDeviceInstanceId accepts number of chars

        Res = !!SetupDiGetDeviceInstanceId (
          DL, const_cast <SP_DEVINFO_DATA *> (&DI),
          Buffer, BufferSize / 2, &DataSize
        );

        DataSize *= 2; // SetupDiGetDeviceInstanceId returns number of chars

      } else if (CompatibleIds) {

        Res = !!SetupDiGetDeviceRegistryProperty (
          DL, const_cast <SP_DEVINFO_DATA *> (&DI),
          SPDRP_COMPATIBLEIDS, &DataType, PBYTE (Buffer), BufferSize, &DataSize
        );

      } else {

        Res = !!SetupDiGetDeviceRegistryProperty (
          DL, const_cast <SP_DEVINFO_DATA *> (&DI),
          SPDRP_HARDWAREID, &DataType, PBYTE (Buffer), BufferSize, &DataSize
        );

      }

      if (Res) {

        if (DataSize != 0) {

          for (PCTSTR p = Buffer; p < Buffer + StrSzToLen (DataSize) && *p; p += wcslen (p) + 1) {

            Debug (5, L"Checking %s Id: %s", IdType, p);

            if (wcscmpi (p, Id) == 0) {

              Debug (7, L"Found %s Id: %s", IdType, p);

              Success = true;

              break;

            }

            if (ByInstanceId) {

              break;

            }

          } // Id compare loop

        }

      } else {

        TranslateSetupApiError ();

        DWORD const Err = GetLastError ();

        if (Err == ERROR_INSUFFICIENT_BUFFER) {

          Debug (5, L"%u bytes required for %s Id", DataSize, IdType);

          if (Buffer != nullptr) {

            delete Buffer;

          }

          Buffer = reinterpret_cast <PTSTR> (new BYTE [DataSize]);

          if (Buffer == nullptr) {

            Error (L"Cannot allocate %u bytes for %s Id buffer", IdType, DataSize);
            break;

          }

          BufferSize = DataSize;

          continue;

        } else {

          if (Err != ERROR_INVALID_DATA) {

            Error (L"Cannot get %s Id (%s)", IdType, ErrorText (Err));

          }

          break;

        }

      } // Res
  
      if (Success || ByInstanceId || CompatibleIds) {
            
        break;

      }

      CompatibleIds = true;

    } // Id type and buffer allocation loop

  } OnlyOnce;

  if (Buffer != nullptr) {

    delete Buffer;
    Buffer = nullptr;

  }

  return Success;

}




      
bool FindDeviceById (HDEVINFO const DL, SP_DEVINFO_DATA & DI, PCTSTR const Id) {

  bool Success = false;

  do {

    bool const InstanceId = (wcschr (Id, L'\\') != nullptr);

    Debug (6, L"Searching for %s Id %s", InstanceId? L"instance" : L"hardware", Id);

    DI.cbSize = sizeof (DI);

    for (UINT i = 0; ; i++) {

      ZeroMemory (&DI, sizeof (DI));
      DI.cbSize = sizeof (DI);

      if (!SetupDiEnumDeviceInfo (DL, i, &DI)) {

        TranslateSetupApiError ();

        if (GetLastError () != ERROR_NO_MORE_ITEMS) {

          Error (L"Cannot get info for device %u (%s)", i, LastErrorText ());

        }

        break;

      }

      if (DoesDeviceIdMatch (DL, DI, Id, InstanceId)) {

        Success = true;

        break;

      }

    } // Device enumeration loop

    if (!Success) {

      Debug (9, L"Device with %s Id %s not found", InstanceId? L"instance" : L"hardware", Id);

    }

  } OnlyOnce;

  return Success;

}





bool IsDevicePresent (PCTSTR const HardwareId) {

  HDEVINFO DL;

  bool Success = false;

  do {

    DL = CreateDL (0, DIGCF_ALLCLASSES | DIGCF_PRESENT);

    if (!DL) {

      break;

    }

    SP_DEVINFO_DATA DI;

    if (FindDeviceById (DL, DI, HardwareId)) {

      Success = true;

    }

  } OnlyOnce;

  DestroyDL (DL);

  return Success;

}





//
// Check if the device needs reboot
// =============================================================================================
//

bool CheckIfRebootRequired (HDEVINFO const DL, SP_DEVINFO_DATA const & DI, bool & RebootRequired) {

  RebootRequired = false;

  bool Success = false;

  do {

    SP_DEVINSTALL_PARAMS DIP;

    DIP.cbSize = sizeof (DIP);

    RebootRequired = false;

    if (!SetupDiGetDeviceInstallParams (DL, const_cast <SP_DEVINFO_DATA *> (&DI), &DIP)) {

      TranslateSetupApiError ();

      Error (L"Cannot get device installer params (%s)", LastErrorText ());

      break;

    }

    if (TestBits (DIP.Flags, DI_NEEDRESTART | DI_NEEDREBOOT)) {

      RebootRequired = true;

    }

  } OnlyOnce;

  return Success;

}





//
// Enable/disable the device
// =============================================================================================
//

bool EnableDevice (HDEVINFO const DL, SP_DEVINFO_DATA const & DI, bool const Enable) {

  bool Success = false;

  do {

    PCTSTR const OpPrefix = Enable? L"en" : L"dis";

    Debug (7, L"%sabling devnode %x", OpPrefix, DI.DevInst);

    SP_PROPCHANGE_PARAMS PCP;

    ZeroMemory (&PCP, sizeof (PCP));

    PCP.ClassInstallHeader.cbSize = sizeof (PCP.ClassInstallHeader);
    PCP.ClassInstallHeader.InstallFunction = DIF_PROPERTYCHANGE;
    PCP.StateChange = DWORD (Enable? DICS_ENABLE : DICS_DISABLE);

    // XP (and maybe 2k too) does not handle DICS_FLAG_CONFIGSPECIFIC properly

    PCP.Scope = static_cast <DWORD> (
      (VersionInfo.dwMajorVersion >= 6)? DICS_FLAG_CONFIGSPECIFIC : DICS_FLAG_GLOBAL
    );

    PCP.HwProfile = 0;

    if (
      !SetupDiSetClassInstallParams (
        DL, const_cast <SP_DEVINFO_DATA *> (&DI), &PCP.ClassInstallHeader,sizeof (PCP)
      )
    ) {

      TranslateSetupApiError ();

      Error (L"Cannot set class installer params (%s)", LastErrorText ());

      break;

    }

    if (!SetupDiCallClassInstaller (DIF_PROPERTYCHANGE, DL, const_cast <SP_DEVINFO_DATA *> (&DI))) {

      TranslateSetupApiError ();

      Error (L"Cannot %sable devnode %x (%s)", OpPrefix, DI.DevInst, LastErrorText ());

      break;

    }

    Debug (8, L"Devnode %x %sabled", DI.DevInst, OpPrefix);

    Success = true;

  } OnlyOnce;

  return Success;

}





//
// Stop device driver with retry
// =============================================================================================
//

bool StopDeviceDriverWithRetry (
  HDEVINFO DL, SP_DEVINFO_DATA const & DI, PCTSTR const DevName, PCTSTR const ServiceId
) {

  Assert (DevName != nullptr && ServiceId != nullptr);

  bool Success = false;

  Debug (7, L"Stopping the driver of device %s before uninstalling it", DevName);

  do {

    UINT ServiceState;

    if (!GetServiceState (ServiceId, ServiceState)) {

      break;

    }

    if (ServiceState == SERVICE_STOPPED) {

      // Driver service is stopped - success

      Debug (8, L"Driver of device %s is already stopped", DevName);

      Success = true;
      break;

    }
        
    bool Err = false;

    for (;;) {

      if (!EnableDevice (DL, DI, false)) {

        Err = true;
        break;

      }

      if (WaitForServiceToStop (ServiceId, 3)) {

        // Driver service is stopped - success

        break;

      }
        
      if (GetLastError () != ERROR_TIMEOUT) {

        Err = true;
        break;

      }

      if (!EnableDevice (DL, DI, true)) {

        Err = true;
        break;

      }

      wchar_t Text [512];

      wsprintf (
        Text,
        L"Cannot stop the driver of device \"%s\"\n"
        L"\n"
        L"If the driver will not be stopped, the device cannot be uninstalled without a reboot.\n"
        L"\n"
        L"Try to close all applications/windows that might use the device, including hidden apps "
        L"having icons in the system tray\n"
        L"\n"
        L"Retry to stop the driver?"
        L"",
        DevName, LastErrorText ()
      );

      int const Reply = MessageBox (HWND_DESKTOP, Text, ProductShortName, MB_YESNO | MB_ICONWARNING);

      if (Reply != IDYES) {

        Err = true;
        break;

      }

    } // loop

    if (Err) {

      break;

    }

    Success = true;

  } OnlyOnce;

  return Success;

}





//
// Remove the device
// =============================================================================================
//

bool RemoveDevice (HDEVINFO const DL, SP_DEVINFO_DATA & DI, PCTSTR const InstanceId) {

  IfDebug (Unreferenced (InstanceId));

  bool Success = false;

  Debug (8, L"DI: node=%x, reserved=%x", DI.DevInst, DI.Reserved);

  do {

    Debug (7, L"Removing device instance \"%s\"", InstanceId);

#if 0

    // It is not recommended to call default DIF_* handlers directly

    if (!SetupDiRemoveDevice (DL, &DI)) {

      TranslateSetupApiError ();

      Error (L"Cannot remove device instance \"%s\" (%s)", InstanceId, LastErrorText ());

      break;

    }

#else

    Debug (6, L"Setting class install params for device node %x", DI.DevInst);

    SP_REMOVEDEVICE_PARAMS RMP;

    RMP.ClassInstallHeader.cbSize = sizeof (RMP.ClassInstallHeader);
    RMP.ClassInstallHeader.InstallFunction = DIF_REMOVE;
    RMP.Scope = DI_REMOVEDEVICE_GLOBAL;
    RMP.HwProfile = 0;

    if (!SetupDiSetClassInstallParams (DL, &DI, &RMP.ClassInstallHeader, sizeof (RMP))) {

      TranslateSetupApiError ();

      Error (L"Cannot set class installer params (%s)", LastErrorText ());

      break;

    }

    Debug (6, L"Removing device node %x", DI.DevInst);

    if (!SetupDiCallClassInstaller (DIF_REMOVE, DL, &DI)) {

      TranslateSetupApiError ();

      Error (L"Cannot remove device instance \"%s\" (%s)", InstanceId, LastErrorText ());

    }

#endif

    Debug (8, L"Removed device instance \"%s\"", InstanceId);

    Success = true;

  } OnlyOnce;

  return Success;

}





//
// Remove driver package from the repository
// =============================================================================================
//

bool RemoveDriverPackage (PCTSTR const InfPath, bool & RestartRecommended) {

  bool Success = false;

  wchar_t const DifxApiDllName [] = L"difxapi.dll";

  HMODULE DifxApi = nullptr;
  
  do {
  
    DifxApi = LoadLibrary (DifxApiDllName);

    if (DifxApi == nullptr) {

      Trace (9, L"Cannot load %s (%s)", DifxApiDllName, LastErrorText ());

      break;

    }

    typedef DWORD WINAPI DriverPackageUninstallW_t (
      __in  PCWSTR DriverPackageInfPath,
      __in  DWORD  Flags,
      __in_opt PCINSTALLERINFO_W pInstallerInfo,
      __out BOOL  *pNeedReboot
    );

    static char const DriverPackageUninstallW_n [] = {
      "DriverPackageUninstallW"
    };

    DriverPackageUninstallW_t * DriverPackageUninstallW_a = (
      (DriverPackageUninstallW_t *) (
        PVOID (GetProcAddress (DifxApi, DriverPackageUninstallW_n))
      )
    );

    if (!DriverPackageUninstallW_a) {

      Trace (9, L"Cannot get address of %S routine", DriverPackageUninstallW_n);

      break;

    }

    DWORD const Flags = DRIVER_PACKAGE_DELETE_FILES;

    Trace (7, L"Calling %S (\"%s\", %#x, nullptr)", DriverPackageUninstallW_n, InfPath, Flags);

    BOOL NeedReboot;

    DWORD const Res = DriverPackageUninstallW_a (
      InfPath, Flags, nullptr, &NeedReboot
    );

    if (Res != NO_ERROR) {

      SetLastError (Res);

      TranslateSetupApiError ();

      Trace (9, L"Cannot uninstall driver package (%s)", LastErrorText ());

      break;

    }

    RestartRecommended = !!NeedReboot;

    Trace (
      8,
      L"Removed driver package for \"%s\", restart %srequired",
      InfPath, RestartRecommended? L"" : L"not "
    );

    Success = true;

  } OnlyOnce;

  if (DifxApi != nullptr) {

    FreeLibrary (DifxApi);

    DifxApi = nullptr;

  }

  return Success;

}





//
// Open service
// =============================================================================================
//

SC_HANDLE OpenSystemService (PCTSTR const Name, UINT const SrvAccess, bool const MustExist) {

  SC_HANDLE SM = 0;
  SC_HANDLE Serv = 0;

  do {

    SM = OpenSCManager (nullptr, nullptr, SC_MANAGER_ENUMERATE_SERVICE);

    if (SM == 0) {

      Error (L"Cannot open Service Manager (%s)", LastErrorText ());

      break;

    }

    Serv = OpenService (SM, Name, SrvAccess);

    if (Serv == 0) {

      if (GetLastError () == ERROR_SERVICE_DOES_NOT_EXIST && !MustExist) {

        Debug (8, L"Service %s does not exist", Name);

      } else {

        Error (L"Cannot open driver service %s (%s)", Name, LastErrorText ());

      }

      break;

    }

  } OnlyOnce;

  if (SM != 0) {

    DWORD const LastErr = GetLastError ();

    CloseServiceHandle (SM);

    SetLastError (LastErr);

  }

  return Serv;

}





//
// Get service state
// =============================================================================================
//

#if 1

bool GetServiceState (PCTSTR const Name, UINT & State) {

  SC_HANDLE Serv = 0;

  bool Success = false;

  do {

    Serv = OpenSystemService (Name, GENERIC_READ);

    if (Serv == 0) {

      Error (L"Cannot open driver service %s (%s)", Name, LastErrorText ());

      break;

    }

    SERVICE_STATUS ServStat;

    if (!QueryServiceStatus (Serv, &ServStat)) {

      Error (L"Cannot query status of service %s (%s)", Name, LastErrorText ());

      break;

    }

    Debug (7, L"Service %s state: %u", Name, ServStat.dwCurrentState);

    State = ServStat.dwCurrentState;

    Success = true;

  } OnlyOnce;

  if (Serv != 0) {

    DWORD const LastErr = GetLastError ();

    CloseServiceHandle (Serv);

    SetLastError (LastErr);

  }

  return Success;

}

#endif





//
// Wait for service stop
// =============================================================================================
//

bool WaitForServiceToStop (PCTSTR const Name, UINT const Sec) {

  Assert (Name != nullptr && Sec != 0);

  SC_HANDLE Serv = 0;

  bool Success = false;

  do {

    Serv = OpenSystemService (Name, GENERIC_READ);

    if (Serv == 0) {

      Error (L"Cannot open driver service %s (%s)", Name, LastErrorText ());

      break;

    }

    SERVICE_STATUS ServStat;

    bool Err = false;

    Debug (7, L"Waiting for service %s to stop...", Name);

    DWORD const StartTime = GetTickCount ();

    do {

      Sleep (200);

      if (!QueryServiceStatus (Serv, &ServStat)) {

        Err = true;

        Error (L"Cannot query status of service %s (%s)", Name, LastErrorText ());

        break;

      }

      Debug (6, L"Service state: %u", ServStat.dwCurrentState);

    } while (ServStat.dwCurrentState != SERVICE_STOPPED && GetTickCount () - StartTime < Sec * 1000);

    if (Err) {

      break;

    }

    if (ServStat.dwCurrentState != SERVICE_STOPPED) {

      Debug (9, L"Service %s isn't stopped after %u s", Name, Sec);

      SetLastError (ERROR_TIMEOUT);

      break;

    }

    Success = true;

  } OnlyOnce;

  if (Serv != 0) {

    DWORD const LastErr = GetLastError ();

    CloseServiceHandle (Serv);

    SetLastError (LastErr);

  }

  return Success;

}





//
// Remove service
// =============================================================================================
//

bool RemoveService (PCTSTR const Name) {

  SC_HANDLE Serv = 0;

  bool Success = false;

  do {

    Serv = OpenSystemService (Name, DELETE, false);

    if (Serv == 0) {

      if (GetLastError () != ERROR_SERVICE_DOES_NOT_EXIST) {

        Error (L"Cannot open driver service %s (%s)", Name, LastErrorText ());

      }

      break;

    }

    if (!DeleteService (Serv)) {

      Warning (L"Cannot delete service %s (%s)", Name, LastErrorText ());

      break;

    }

    Debug (7, L"Service %s deleted", Name);

    Success = true;

  } OnlyOnce;

  if (Serv != 0) {

    DWORD const LastErr = GetLastError ();

    CloseServiceHandle (Serv);

    SetLastError (LastErr);

  }

  return Success;

}
